from pathlib import Path

from m.core import Bad, Res, issue, one_of

from .config import Config, read_config
from .docker.env import MEnvDocker
from .docker.filenames import FileNames


def _write_blueprints(
    config: Config,
    *,
    m_tag: str,
    cache_from_pr: str,
    update_makefile: bool,
    update_workflow: bool,
) -> Res[None]:
    m_dir = config.m_dir
    docker_config = config.docker_config
    if not docker_config:
        return issue(
            'missing docker_config in m file',
            context={'m_dir': m_dir},
        )

    env_docker = MEnvDocker(
        m_tag=m_tag,
        cache_from_pr=cache_from_pr,
        base_path=docker_config.base_path,
        registry=docker_config.docker_registry,
        multi_arch=bool(docker_config.architectures),
    )
    files = FileNames.create_instance(m_dir)
    if update_makefile:
        update_res = docker_config.update_makefile(files)
        if isinstance(update_res, Bad):
            return Bad(update_res.value)
    if update_workflow:
        update_res = docker_config.update_github_workflow(files)
        if isinstance(update_res, Bad):
            return Bad(update_res.value)
    return docker_config.write_blueprints(m_dir, env_docker)


def write_blueprints(
    m_dir: str,
    *,
    m_tag: str,
    cache_from_pr: str,
    update_makefile: bool = False,
    update_workflow: bool = False,
) -> Res[None]:
    """Write a file with the M environment variables.

    Args:
        m_dir: The directory with the m configuration.
        m_tag: The unique identifier for the build.
        cache_from_pr: The pr number to attempt to pull cache from.
        update_makefile: If true, updates the `Makefile`.
        update_workflow: If true, updates the github workflow.

    Returns:
        An issue or the m environment instance.
    """
    blueprints_dir = Path(f'{m_dir}/.m/blueprints')
    if not blueprints_dir.exists():
        Path(f'{m_dir}/.m/blueprints/local').mkdir(parents=True)
        Path(f'{m_dir}/.m/blueprints/ci').mkdir(parents=True)
    return one_of(lambda: [
        None
        for config in read_config(m_dir)
        for _ in _write_blueprints(
            config,
            m_tag=m_tag,
            cache_from_pr=cache_from_pr,
            update_makefile=update_makefile,
            update_workflow=update_workflow,
        )
    ])



from .disable import color_disabled


class Color:
    """A class for terminal color codes."""

    bold = '\033[1m'
    blue = '\033[94m'
    aqua = '\033[96m'
    white = '\033[97m'
    green = '\033[92m'
    orange = '\033[38;5;172m'
    yellow = '\033[93m'
    gray = '\033[90m'
    red = '\033[91m'
    purple = '\033[95m'
    bold_white = bold + white
    bold_blue = bold + blue
    bold_aqua = bold + aqua
    bold_green = bold + green
    bold_orange = bold + orange
    bold_yellow = bold + yellow
    bold_gray = bold + gray
    bold_red = bold + red
    bold_purple = bold + purple
    end = '\033[0m'


color_dict = {
    color_name: color_value
    for color_name, color_value in Color.__dict__.items()
    if color_name[0] != '_'
}
no_color_dict = {color_name: '' for color_name in color_dict}


def color(*args: str, auto_end: bool = True) -> str:
    r"""Color a message.

    Format the arguments by replacing the colors in the Colors class.
    For instance::

        color('{blue}Hello there{end}\n{yellow}WARNING')

    Args:
        args: Strings to color
        auto_end: Add a string to end coloring

    Returns:
        A formatted message.
    """
    no_color = color_disabled()
    color_map = no_color_dict if no_color else color_dict
    end = '' if no_color or not auto_end else Color.end
    msg_list = [msg.format(**color_map) for msg in args]
    return ''.join(msg_list) + end



from importlib import import_module
from typing import Any

from m.core import Good, Res, issue


def import_attr(attr_path: str) -> Res[Any]:
    """Get an attribute from a module.

    The `attr_path` should be a fully qualified path to the attribute. For
    example `import_attr('m.core.one_of')` will attempt to retrieve a handle
    on `one_of`.

    Args:
        attr_path: The path to the attribute.

    Returns:
        A `Good` containing the attribute or an issue.
    """
    module_parts = attr_path.split('.')
    module_name = '.'.join(module_parts[:-1])
    item_name = module_parts[-1]
    try:
        module = import_module(module_name)
    except Exception as ex:
        return issue(
            'import_module_failure',
            cause=ex,
            context={'attr_path': attr_path, 'module_name': module_name},
        )
    module_dict = module.__dict__
    if item_name not in module_dict:
        return issue(
            'missing_attribute',
            context={'module': str(module), 'missing_attribute': item_name},
        )
    return Good(module_dict[item_name])



"""
This module is provided to help with testing.

The following external modules are expected to be available:

- [pytest][]
- [pytest-mock](https://pytest-mock.readthedocs.io/en/latest/)
"""
import os
import runpy
import sys
from io import StringIO
from typing import Any

import pytest
from m.core import Good
from pytest_mock import MockerFixture

from .testing import ActionStepTestCase


def run_action_step(
    mocker: MockerFixture,
    *,
    py_file: str,
    exit_code: int,
    env_vars: dict[str, str],
    file_write_side_effect: Any | None = None,
) -> tuple[str, str, dict[str, str]]:
    """Execute an action step in a test.

    This function expects the inputs to the script to be provided via environment
    variables of the form `INPUT_[SOME_NAME]`. The script will write the outputs
    to the file `FAKE_GITHUB_OUTPUT.txt`. We can verify the contents of the file
    by looking at the 3rd output from the function. This is a dictionary mapping
    file names to contents. Please note that this testing function mocks
    [m.core.rw.write_file][] to obtain the file contents.

    Args:
        mocker: A reference to the pytest `MockerFixture`.
        py_file: The full path to the file that Github Actions will run.
        exit_code: The expected exit code of the action. `0` means all is good.
        env_vars: A dictionary of the environment variables that the action will
            receive.
        file_write_side_effect: This can be provided if we need to modify the
            behavior of [m.core.rw.write_file][]. This is useful if we want to
            test cases in which a file failed to write.

    Returns:
        The standard out, standard error, and files written by [m.core.rw.write_file][].
    """
    mocker.patch.dict(
        os.environ,
        {
            'NO_COLOR': 'true',
            **env_vars,
            'GITHUB_OUTPUT': 'FAKE_GITHUB_OUTPUT.txt',
        },
        clear=True,
    )

    std_out = StringIO()
    std_err = StringIO()
    mocker.patch.object(sys, 'stdout', std_out)
    mocker.patch.object(sys, 'stderr', std_err)
    file_write_mock = mocker.patch('m.core.rw.write_file')
    file_write_mock.side_effect = file_write_side_effect or [Good(0)]

    prog = None
    with pytest.raises(SystemExit) as prog_block:
        prog = prog_block
        runpy.run_path(py_file, {}, '__main__')

    # Would be nice to be able to reset via a the mocker
    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__
    assert prog is not None  # noqa: S101 - to be used in testing
    file_writes = {
        call.args[0]: call.args[1]
        for call in file_write_mock.call_args_list
    }

    # next block should not be covered by coverage, we have this as a utility
    # to help us write tests.
    prog_code = prog.value.code
    if prog_code != exit_code:  # pragma: no cover
        print(std_out.getvalue(), file=sys.stdout)  # noqa: WPS421
        print(std_err.getvalue(), file=sys.stderr)  # noqa: WPS421
    assert prog_code == exit_code   # noqa: S101 - to be used in testing
    return std_out.getvalue(), std_err.getvalue(), file_writes


def run_action_test_case(
    mocker: MockerFixture,
    tcase: ActionStepTestCase,
) -> None:
    """Execute an action step test case.

    This is a commodity wrapper to help us run the action tests case. If we need
    more control over the assertions we can then copy and modify the implementation.

    Args:
        mocker: A reference to the pytest `MockerFixture`.
        tcase: The test case.
    """
    stdout, stderr, file_writes = run_action_step(
        mocker,
        py_file=tcase.py_file,
        exit_code=tcase.exit_code,
        env_vars=tcase.inputs,
        file_write_side_effect=tcase.file_write_side_effect,
    )
    assert stdout == tcase.expected_stdout  # noqa: S101
    if tcase.errors:
        for error in tcase.errors:
            assert error in stderr  # noqa: S101

    if tcase.exit_code == 0:
        assert 'FAKE_GITHUB_OUTPUT.txt' in file_writes  # noqa: S101
        gh_output = file_writes['FAKE_GITHUB_OUTPUT.txt']
        assert gh_output == '\n'.join(tcase.outputs)  # noqa: S101



from m.cli import Arg, BaseModel, command, env_var_or_empty, run_main


class Arguments(BaseModel):
    """Create the `[m_dir]/.m/blueprints` directory.

    This will add shell scripts to build the specified docker images as
    stated in the `[m_dir]/m.yaml` file.
    """

    m_dir: str = Arg(
        default='m',
        help='m project directory',
        positional=True,
    )
    m_tag: str = Arg(
        default='M_TAG',
        validator=env_var_or_empty,
        help='unique version to use for all the images',
    )
    cache_from_pr: str = Arg(
        default='M_CACHE_FROM_PR',
        validator=env_var_or_empty,
        help='pull request number to attempt to use as cache',
    )
    skip_makefile: bool = Arg(
        default=False,
        help='do not update Makefile',
    )
    skip_workflow: bool = Arg(
        default=False,
        help='do not update github workflow',
    )


@command(
    help='create docker images blueprints',
    model=Arguments,
)
def run(arg: Arguments) -> int:
    from m.ci.m_blueprints import write_blueprints

    return run_main(
        lambda: write_blueprints(
            arg.m_dir,
            m_tag=arg.m_tag,
            cache_from_pr=arg.cache_from_pr,
            update_makefile=not arg.skip_makefile,
            update_workflow=not arg.skip_workflow,
        ),
        result_handler=lambda _: None,
    )



from m.cli import Arg, BaseModel, command, run_main
from m.core.io import env


class Arguments(BaseModel):
    r"""Retrieve the information required for continuous integration.

    example::

        $ m github ci \
            --owner jmlopez-rod \
            --repo pysync \
            --sha 4538b2a2556efcbdfc1e7df80c4f71ade45f3958 \
            --pr 1 \
            --include-release | m json
        {
        "commit": {
            "associatedPullRequests": {
            "nodes": [
                {
                "author": {
                    "login": "jmlopez-rod",
                    "avatarUrl": "https://avatars.githubusercontent.com/...",
                    "email": ""
                },
        ...

    NOTE: Use the --merge-commit flag if you are providing a sha from
    github actions.
    """

    owner: str = Arg(
        default=env('GITHUB_REPOSITORY_OWNER'),
        help='repo owner',
    )
    repo: str = Arg(
        help='repo name',
        required=True,
    )
    sha: str = Arg(
        help='commit sha',
        required=True,
    )
    pr: int | None = Arg(help='pull request number')
    file_count: int = Arg(
        default=10,
        help='max number of files to retrieve',
    )
    include_release: bool = Arg(
        default=False,
        help='include the last release information',
    )
    merge_commit: bool = Arg(
        default=False,
        help='set if the sha is a merge commit sha (from github)',
    )


@command(
    help='continuous integration information',
    model=Arguments,
)
def run(arg: Arguments, arg_ns) -> int:
    from m.github.ci import CommitInfo, get_raw_ci_run_info
    return run_main(lambda: get_raw_ci_run_info(
        arg_ns.token,
        CommitInfo(owner=arg.owner, repo=arg.repo, sha=arg.sha),
        arg.pr,
        arg.file_count,
        arg.include_release,
        get_sha=arg.merge_commit,
    ))



import time

from pydantic import BaseModel

from .config import Config


class VersionInputs(BaseModel):
    """Container with information to generate a version."""

    # '0.0.0-' or 'a.b.c-' or '': Necessary for pr builds
    version_prefix: str

    # the version in the configuration
    version: str

    # unique number for each run in CI environment
    run_id: str | None

    # current commit sha
    sha: str

    pr_number: int | None

    # The build git branch
    branch: str

    is_release: bool
    is_release_pr: bool
    is_hotfix_pr: bool


def build_m_tag(ver_input: VersionInputs, config: Config) -> str:
    """Build an valid "M_TAG".

    Args:
        ver_input: The inputs to create the tag.
        config: The m configuration.

    Returns:
        An m tag that can be used to version npm and docker packages.
    """
    prefix = '' if config.uses_free_flow() else f'{ver_input.version_prefix}-'
    pr_number = ver_input.pr_number
    run_id = ver_input.run_id
    if not run_id:
        return f'{prefix}local.{ver_input.sha}'
    if ver_input.is_release:
        return ver_input.version
    if pr_number:
        nprefix = ''
        if ver_input.is_release_pr:
            nprefix = 'rc'
        elif ver_input.is_hotfix_pr:
            nprefix = 'hotfix'
        if nprefix:
            return f'{ver_input.version}-{nprefix}{pr_number}.b{run_id}'
        return f'{prefix}pr{pr_number}.b{run_id}'
    return f'{prefix}{ver_input.branch}.b{run_id}'


def build_py_tag(ver_input: VersionInputs, config: Config) -> str:
    """Build a valid python version.

    The configuration object is provided since python does not accept using
    names of branches in the versions. For this reason we will map

    Args:
        ver_input: The inputs to create the version.
        config: The m configuration.

    Returns:
        A valid python tag.
    """
    prefix = ver_input.version_prefix
    pr_number = ver_input.pr_number
    run_id = ver_input.run_id
    now = int(time.time())
    if not run_id:
        return f'{prefix}a0+b{now}'
    if ver_input.is_release:
        return ver_input.version
    if pr_number:
        nprefix = ''
        if ver_input.is_release_pr or ver_input.is_hotfix_pr:
            nprefix = 'rc'
        if nprefix:
            return f'{ver_input.version}{nprefix}{pr_number}.dev{run_id}'
        return f'{prefix}b{pr_number}.dev{run_id}'
    nprefix = _get_py_branch_prefix(config, ver_input.branch)
    return f'{prefix}{nprefix}.dev{run_id}'


def _get_py_branch_prefix(config: Config, branch: str) -> str:
    master_branch = config.get_master_branch()
    develop_branch = (
        config.git_flow.develop_branch
        if config.uses_git_flow()
        else 'develop'
    )
    if branch == master_branch:
        return 'rc0'
    if branch == develop_branch:
        return 'b0'
    # letting a0 denote any branch other than the ones for master/develop
    return 'a0'



from typing import cast

from ..types import AnyMap, FuncArgs


def should_handle(extras: AnyMap) -> bool:
    """Handle the proxy field.

    Args:
        extras: A dictionary with information for a cli argument.

    Returns:
        True if it should handle the field as a proxy to the parser.
    """
    return extras.get('proxy', None) is not None


def handle_field(extras: AnyMap) -> FuncArgs:
    """Pass the `proxy` property in the field as is.

    Args:
        extras: A dictionary representation of the field.

    Returns:
        Function arguments for the parser `add_argument` method.
    """
    # proxy field is expected to be a FuncArgs instance.
    return cast(FuncArgs, extras['proxy'])



from m.cli import Arg, BaseModel, command, run_main


class Arguments(BaseModel):
    """Supplement to `m github release`.

    This command needs the `git` cli and it is meant to be run after
    `m github release` to create a major and minor release tags.


    example::

        $ m git tag_release --version 1.2.3


    It will create or update the following tags::

        -v1
        -v1.2

    Note that the tags may be moved with each release to point to the latest
    release. This is done by deleting the tag and creating it again.
    """

    version: str = Arg(
        help='version to create tags from',
        required=True,
    )
    sha: str = Arg(
        default='',
        help='sha to tag',
    )
    major_only: bool = Arg(
        default=False,
        help='only create major tag',
    )


@command(
    help='tag github releases',
    model=Arguments,
)
def run(arg: Arguments) -> int:
    from m import git
    return run_main(
        lambda: git.tag_release(arg.version, arg.sha, major_only=arg.major_only),
        result_handler=print,
    )



from m.cli import ArgProxy, BaseModel, command, env_var, run_main


class Arguments(BaseModel):
    """Start the release process for a hotfix.

    A hotfix should always be done in the master branch.

    It may also require input from the developer to proceed with
    certain operations.
    """

    github_token: str = ArgProxy(
        '--github-token',
        default='GITHUB_TOKEN',
        type=env_var,
        help='Github personal access token with repo scope',
    )


@command(
    help='start the release process for a hotfix',
    model=Arguments,
)
def run(arg: Arguments):
    from m.ci.start_release import start_release

    return run_main(lambda: start_release(arg.github_token, hotfix=True))



from typing import cast

from pydantic.fields import FieldInfo

from ..misc import argument_description, argument_name
from ..types import MISSING, AnyMap, FuncArgs


def handle_field(name: str, field: FieldInfo) -> FuncArgs:
    """Handle a standard field as an argument.

    Args:
        name: The name of the argument.
        field: A dictionary representation of the field.

    Returns:
        Function arguments for the parser `add_argument` method.
    """
    extras = cast(dict, field.json_schema_extra or {})
    default = field.default
    validator = extras.get('validator', None)
    required = extras.get('required', False)
    aliases = cast(list[str], extras.get('aliases', None))

    arg_default = default if default is MISSING else repr(default)
    if 'env_var' in repr(validator):
        # special case - do not repr
        arg_default = f'env.{default}'
    args: AnyMap = {
        'help': argument_description(field.description or '', arg_default),
        'required': required,
        'type': str,
    }

    if default is not MISSING:
        args['default'] = default

    if validator:
        args['type'] = validator

    names = [argument_name(name)]
    if aliases:
        names = [argument_name(alias) for alias in aliases]

    return FuncArgs(args=names, kwargs=args)



from m.core import io as mio

from .providers.github import tool as gh_tool
from .providers.local import tool as local_tool
from .providers.teamcity import tool as tc_tool
from .types import ProviderModule


def get_ci_tool() -> ProviderModule:
    """Return the current CI Tool based on the environment variables.

    Returns:
        A `ProviderModule` instance with methods to provide messages in
        a CI environment.
    """
    env = mio.env
    if env('GITHUB_ACTIONS'):
        return gh_tool
    if env('TC') or env('TEAMCITY'):
        return tc_tool
    return local_tool



from typing import Any

from m.core import Issue, OneOf, http


def request(
    token: str,
    endpoint: str,
    method: http.HttpMethod = http.HttpMethod.get,
    dict_data: object | None = None,
) -> OneOf[Issue, Any]:
    """Make an api request to github.

    See::

        - https://docs.github.com/en/rest/overview/resources-in-the-rest-api
        - https://docs.github.com/en/rest/overview/endpoints-available-for-github-apps

    Args:
        token: A github personal access token.
        endpoint: A github api endpoint.
        method: The http method to use. (default 'GET')
        dict_data: A payload if the method if `POST` or `GET`.

    Returns:
        A response from Github.
    """
    url = f'https://api.github.com{endpoint}'
    headers = {'authorization': f'Bearer {token}'}
    return http.fetch_json(url, headers, method, dict_data)



from .gh_workflow_single import TemplateVars as DefaultTemplateVars
from .gh_workflow_single import Workflow as DefaultWorkflow

TEMPLATE = """\
# AUTOGENERATED FILE - Do not update by hand!
# Edit {m_dir}/m.yaml then run
# `m blueprints` to update
name: m

on:
  workflow_call:
    inputs:
      m-tag:
        type: string
        description: The unique version to use for all the images
        required: true
      cache-from-pr:
        type: string
        description: The pull request number to attempt to use as cache.
        required: true{extra_inputs}

permissions: write-all

env:{global_env}

jobs:
  blueprints:
    runs-on: {default_runner}
    outputs:
      image-names: ${{{{ steps.m-blueprints.outputs.image-names }}}}
      image-tags: ${{{{ steps.m-blueprints.outputs.image-tags }}}}
    steps:
      - name: checkout
        uses: actions/checkout@v4
      - name: install-m
        run: pip install jmlopez-m
      - name: m-blueprints
        id: m-blueprints
        run: |-
          m blueprints --skip-makefile --skip-workflow {m_dir}
          {{
            echo "image-names=$(< {ci_dir}/_image-names.json)"
            echo "image-tags=$(< {ci_dir}/_image-tags.json)"
          }} >> $GITHUB_OUTPUT
      - name: archive
        uses: actions/upload-artifact@v4
        with:
          name: m-blueprints
          path: {ci_dir}

  build:
    needs: blueprints
    strategy:
      fail-fast: false
      matrix:
        include:{build_architectures}
    runs-on: ${{{{ matrix.os }}}}{container}
    steps:
      - name: checkout
        uses: actions/checkout@v4
      - name: restore-m-blueprints
        uses: actions/download-artifact@v4
        with:
          name: m-blueprints
          path: {ci_dir}
      - name: chmod
        run: chmod +x {ci_dir}/*.sh
      - {docker_login}
      {build_steps}

  manifest:
    runs-on: {default_runner}
    needs: [blueprints, build]
    strategy:{manifest_strategy_options}
      matrix:
        image-name: ${{{{ fromJSON(needs.blueprints.outputs.image-names) }}}}
        image-tag: ${{{{ fromJSON(needs.blueprints.outputs.image-tags) }}}}
    steps:
      - {docker_login}
      - name: create
        run: {create_manifest}
      - name: push
        run: {push_manifest}
"""


class TemplateVars(DefaultTemplateVars):
    """Template variables."""

    build_architectures: str

    create_manifest: str

    push_manifest: str

    manifest_strategy_options: str


def _indent(text: str, num: int) -> str:
    spaces = '  ' * num
    return text.replace('\n', f'\n{spaces}', -1)


class Workflow(DefaultWorkflow):
    """Helper class to write the `m` workflow."""

    architectures: dict[str, str | list[str]]

    max_parallel_manifests: int | None

    def build_architectures(self: 'Workflow') -> str:
        """Generate a github action str with the build architectures.

        Returns:
            A string to add to the Github workflow.
        """
        arch_strs = '\n'.join([
            f'- arch: {arch}\n  os: {os}'
            for arch, os in self.architectures.items()
        ])
        return _indent(f'\n{arch_strs}', 5)

    def create_manifest_str(self: 'Workflow') -> str:
        """Generate the script to create the manifest.

        Returns:
            The manifest create script.
        """
        cmd = 'docker manifest create'
        registry = self.docker_registry
        image = '${{ matrix.image-name }}'
        tag = '${{ matrix.image-tag }}'
        m_tag = '${{ inputs.m-tag }}'
        lines = [f'{cmd} {registry}/{image}:{tag}']
        for arch in self.architectures:
            lines.append(f'  {registry}/{arch}-{image}:{m_tag}')
        # wants it be a raw string but i need a new line after `\`
        full_cmd = ' \\\n'.join(lines)   # noqa: WPS342
        return _indent(f'|-\n{full_cmd}', 5)

    def push_manifest_str(self: 'Workflow') -> str:
        """Generate the script to run to push the manifest.

        Returns:
            The command to push.
        """
        cmd = 'docker manifest push'
        registry = self.docker_registry
        image = '${{ matrix.image-name }}'
        tag = '${{ matrix.image-tag }}'
        full_cmd = f'|-\n{cmd} {registry}/{image}:{tag}'
        return _indent(full_cmd, 5)

    def manifest_strategy_options_str(self: 'Workflow') -> str:
        """Generate the strategy options for the manifest job.

        Returns:
            The strategy options.
        """
        options = ''
        if self.max_parallel_manifests:
            options = f'\n      max-parallel: {self.max_parallel_manifests}'
        return options

    def __str__(self: 'Workflow') -> str:
        """Stringify the workflow file.

        Returns:
            The github workflow.
        """
        template_vars = TemplateVars(
            m_dir=self.m_dir,
            default_runner=self.default_runner,
            global_env=self.global_env_str(),
            extra_inputs=self.extra_inputs_str(),
            ci_dir=self.ci_dir,
            build_architectures=self.build_architectures(),
            docker_login=self.docker_login_str(),
            build_steps=self.build_steps_str(),
            create_manifest=self.create_manifest_str(),
            push_manifest=self.push_manifest_str(),
            manifest_strategy_options=self.manifest_strategy_options_str(),
            container=self.container_str(),
        )
        return TEMPLATE.format(**template_vars.model_dump())



from m.cli import Arg, BaseModel, command


class Arguments(BaseModel):
    """Close a block.

    When a block is closed, all its inner blocks are closed automatically.
    Not all CI Tools support nesting.
    """

    name: str = Arg(
        help='block name to close',
        positional=True,
        required=True,
    )


@command(
    help='close a block',
    model=Arguments,
)
def run(arg: Arguments):
    from m.log import Logger

    logger = Logger('m.cli')
    logger.close_block(arg.name)
    return 0



import re
from typing import Dict, List

from ....core import Good, Issue, OneOf
from ..core.types import FileReport, Violation


def read_payload(payload: str) -> OneOf[Issue, List[FileReport]]:
    """Transform a pycodestyle payload to a list of `FileReport` instances.

    This function can be used with flake8 and other tools that emit similar
    output.

    Args:
        payload: The raw payload from pycodestyle.

    Returns:
        A `OneOf` containing an `Issue` or a list of `FileReport` instances.
    """
    regex = r'(.*):(\d+):(\d+): (\w+) (.*)'
    report: Dict[str, List[Violation]] = {}
    for line in payload.splitlines():
        match = re.match(regex, line)
        if match:
            group = match.groups()
            violation = Violation(
                file_path=group[0],
                line=int(group[1]),
                column=int(group[2]),
                rule_id=group[3],
                message=group[4],
            )
            if violation.file_path not in report:
                report[violation.file_path] = []
            report[violation.file_path].append(violation)
    return Good([
        FileReport(file_path=name, violations=violations)
        for name, violations in report.items()
    ])



from typing import Any, Mapping

from m.core import Good, Issue, OneOf, http, issue, one_of

from ..request import request

HttpMethod = http.HttpMethod


def _filter_data(dict_data: Mapping[str, Any]) -> OneOf[Issue, Any]:
    if dict_data.get('errors'):
        return issue(
            'github graphql errors',
            context={'response': dict_data},
        )
    if dict_data.get('data'):
        return Good(dict_data['data'])
    return issue(
        'github response missing data field',
        context={'response': dict_data},
    )


def graphql(
    token: str,
    query: str,
    variables: Mapping[str, Any],
) -> OneOf[Issue, Any]:
    """Make a request to Github's graphql API.

    https://docs.github.com/en/graphql/guides/forming-calls-with-graphql

    Args:
        token: A github PAT.
        query: A graphql query.
        variables: The variables to use in the query.

    Returns:
        The Github response.
    """
    payload = {'query': query, 'variables': variables or {}}
    return one_of(
        lambda: [
            payload
            for res in request(token, '/graphql', HttpMethod.post, payload)
            for payload in _filter_data(res)
        ],
    )



import logging

from m.core import Good, Issue, OneOf
from m.log.misc import default_record_fmt, format_context, format_location

from ..types import EnvVars, Message, ProviderModule

REPLACEMENTS = (
    ('|', '||'),
    ("'", "|'"),
    ('[', '|['),
    (']', '|]'),
    ('\n', '|n'),
)


def escape_msg(msg: str) -> str:
    """Escapes characters so Teamcity can print correctly.

    https://www.jetbrains.com/help/teamcity/build-script-interaction-with-teamcity.html#BuildScriptInteractionwithTeamCity-Escapedvalues

    Args:
        msg: The message to process.

    Returns:
        A message that can be provided to Teamcity.
    """
    message = msg
    for target, replacement in REPLACEMENTS:
        message = message.replace(target, replacement)
    return message


def _escape(name: str, description: str) -> str:
    desc = escape_msg(description)
    return f"{name}='{desc}'"


def _tc(cmd: str, postfix, **kwargs) -> str:
    key_items = ' '.join(
        [_escape(name, desc) for name, desc in kwargs.items()],
    ).strip()
    return f'##teamcity[{cmd} {key_items}]{postfix}'


def env_vars() -> OneOf[Issue, EnvVars]:
    """Read basic environment variables from Teamcity.

    Returns:
        An EnvVar instance.
    """
    # WIP: Need to map other variables here. Not willing to do
    # at the moment since I'm using Github.
    return Good(EnvVars(ci_env=True))


def _fmt_open(record_dict: dict) -> str:
    open_b = record_dict.get('open_block')
    if open_b:
        name, desc = open_b
        return _tc('blockOpened', postfix='', name=name, description=desc)
    return ''


def log_format(
    _formatter: logging.Formatter,
    record: logging.LogRecord,
    show_traceback: bool,
    debug_python: bool,
) -> str:
    """Format a log record using the functions provided in this module.

    This function is meant to be used by a log formatter. See more info

    See::
        https://www.jetbrains.com/help/teamcity/build-script-interaction-with-teamcity.html#BuildScriptInteractionwithTeamCity-BlocksofServiceMessages

    Args:
        _formatter: A log formatter instance.
        record: A log record.
        show_traceback: If true, display the python stack trace.
        debug_python: If true, display the location of the record's origin.

    Returns:
        A formatted string.
    """
    record_dict = record.__dict__

    open_b = _fmt_open(record_dict)
    if open_b:
        return open_b

    close_b = record_dict.get('close_block')
    if close_b:
        return _tc('blockClosed', postfix='', name=close_b)

    indent_padding = 2 if record.levelname in {'WARNING', 'ERROR'} else 3
    indent = len(record.levelname) + indent_padding

    context = format_context(record, indent, show_traceback=show_traceback)
    if not record.msg:
        return context[1:]

    ci_info = record_dict.get('ci_info', Message(msg=record.msg))
    msg_info = format_location([ci_info.file, ci_info.line, ci_info.col])
    msg_info = f'{msg_info}:' if msg_info else ''

    loc = (
        format_location([record.pathname, f'{record.lineno}'])
        if debug_python
        else ''
    )

    msg = f'{loc}{msg_info} {record.msg}'.lstrip()
    if record.levelname == 'WARNING':
        msg = _tc(
            'message',
            status='WARNING',
            text=msg,
            postfix=context,
        )
    elif record.levelname == 'ERROR':
        msg = _tc('buildProblem', postfix=context, description=msg)
    else:
        msg = default_record_fmt(record, '', f'{msg_info}{loc}', context)
    return msg


tool = ProviderModule(
    ci=True,
    env_vars=env_vars,
    formatter=log_format,
)



import argparse as ap
from glob import iglob
from importlib import import_module
from os import path as pth
from typing import Callable, cast

from ..args import Meta
from .types import (
    CliCommands,
    CliSubcommands,
    CommandModule,
    CommandModuleMap,
    DecoratedRunFunction,
    MetaModule,
)

# Used as a default meta for the root command.
default_root_meta = MetaModule(
    meta=Meta(
        help='execute commands',
        description="""
            See the help for each of the following supported commands.
        """,
    ),
)


# Used as default meta for all Subcommands.
default_meta = MetaModule(
    meta=Meta(
        help='WIP - help message not set',
        description='WIP - description not set',
    ),
)


def _get_command_module(name: str) -> CommandModule | None:
    cmd_mod = import_module(name).__dict__
    run_func = cmd_mod.get('run')
    if run_func:
        return CommandModule(run=run_func)
    return None


def _get_meta_module(name: str, default: MetaModule) -> MetaModule:
    meta_mod = import_module(name).__dict__
    return MetaModule(
        meta=meta_mod.get('meta', default.meta),
        add_arguments=meta_mod.get('add_arguments', default.add_arguments),
    )


def get_command_modules(commands_module: str) -> dict[str, CommandModule]:
    """Create a dictionary of a file name to its module.

    These modules should define a `run` function.

    Args:
        commands_module: The full module resolution. For instance
            `m.cli.commands`.

    Returns:
        A dictionary of command modules.
    """
    cmd_module = import_module(commands_module)
    dir_name = pth.split(cmd_module.__file__ or '')[0]
    mod_names = list(iglob(f'{dir_name}/*.py'))
    mod = {}
    for mod_name in mod_names:
        name = pth.split(mod_name)[1][:-3]
        cmd_mod = _get_command_module(f'{commands_module}.{name}')
        if cmd_mod:
            mod[name] = cmd_mod
    return mod


def import_cli_commands(commands_module: str) -> CliCommands:
    """Gather the commands and subcommands for the cli.

    Args:
        commands_module: module containing all the commands.

    Returns:
        An instance of CliCommands gathered from the commands_module.
    """
    cmd_module = import_module(commands_module)
    if hasattr(cmd_module, 'create_cli_commands'):
        # No extra checks for now, if a module has this we assume that
        # it has the right signature.
        return cast(CliCommands, cmd_module.create_cli_commands())

    commands: dict[str, CommandModule | CliSubcommands] = {}
    root_cmd = get_command_modules(commands_module)

    for key, cmd_mod in root_cmd.items():
        commands[key] = cmd_mod

    root = pth.split(cmd_module.__file__ or '')[0]
    for cmd_name in iglob(f'{root}/*'):
        if cmd_name.endswith('.py') or cmd_name.endswith('__'):
            continue
        name = pth.split(cmd_name)[1]
        commands[name] = CliSubcommands(
            meta=_get_meta_module(f'{commands_module}.{name}', default_meta),
            subcommands=get_command_modules(f'{commands_module}.{name}'),
        )

    return CliCommands(
        meta=_get_meta_module(commands_module, default_root_meta),
        commands=commands,
    )


def subcommands(
    meta: MetaModule | None = None,
    **commands: DecoratedRunFunction,
) -> CliSubcommands:
    """Create a CliSubcommands instance.

    Args:
        meta: The meta for the command group.
        commands: The commands the group.

    Returns:
        An instance of CliSubcommands.
    """
    return CliSubcommands(
        meta=meta,
        subcommands={
            cmd_name: CommandModule(run=cmd_item)
            for cmd_name, cmd_item in commands.items()
        },
    )


def cli_commands(
    root_meta: MetaModule | None = None,
    **commands: DecoratedRunFunction | CliSubcommands,
) -> CliCommands:
    """Create a CliCommands instance.

    Args:
        root_meta: The meta for the root command.
        commands: The commands and subcommands for the cli.

    Returns:
        An instance of CliCommands.
    """
    root_meta = root_meta or default_root_meta
    return CliCommands(
        meta=root_meta,
        commands={
            cmd_name: (
                cmd_item
                if isinstance(cmd_item, CliSubcommands)
                else CommandModule(run=cmd_item)
            )
            for cmd_name, cmd_item in commands.items()
        },
    )


def command_group(
    *,
    help: str,   # noqa: WPS125
    description: str,
    add_arguments: Callable[[ap.ArgumentParser], None] | None = None,
) -> MetaModule:
    """Create an instance of a MetaModule.

    Named `cmd_group` since it is used to describe the group of commands.

    Args:
        help: quick help describing the module.
        description: Detailed description about the module.
        add_arguments: Optional function to handle the argparse instance.

    Returns:
        An instance of a MetaModule.
    """
    return MetaModule(
        meta=Meta(help=help, description=description),
        add_arguments=add_arguments,
    )


# A function to resolve merge conflicts between two Subcommands.
SubCmdResolution = Callable[[CommandModuleMap, CommandModuleMap], CommandModuleMap]


def merge_cli_commands(
    base: CliCommands,
    overrides: CliCommands,
    **resolutions: SubCmdResolution,
) -> CliCommands:
    """Merge two CliCommands instances.

    Resolutions may be provided to resolve merge conflicts between two
    subcommands.

    Args:
        base: The base CliCommands instance.
        overrides: The overrides CliCommands instance.
        resolutions: A dictionary of resolutions for subcommands.

    Returns:
        A new CliCommands instance.
    """
    commands: dict[str, CommandModule | CliSubcommands] = {**base.commands}
    for cmd_name, cmd_item in overrides.commands.items():
        resolution = resolutions.get(cmd_name)
        commands[cmd_name] = _get_new_command(cmd_name, commands, cmd_item, resolution)

    return CliCommands(
        meta=base.meta or overrides.meta,
        commands=commands,
    )


def _get_new_command(
    name: str,
    commands: dict[str, CommandModule | CliSubcommands],
    new_command_item: CliSubcommands | CommandModule,
    resolution: SubCmdResolution | None = None,
) -> CliSubcommands | CommandModule:
    if name in commands:
        if isinstance(new_command_item, CliSubcommands):
            current_command_item = commands.get(name)
            if isinstance(current_command_item, CliSubcommands):
                return CliSubcommands(
                    meta=new_command_item.meta or current_command_item.meta,
                    subcommands=_merge_subcommands(
                        current_command_item,
                        new_command_item,
                        resolution,
                    ),
                )
    return new_command_item


def _merge_subcommands(
    current_command_item: CliSubcommands,
    new_command_item: CliSubcommands,
    resolution: SubCmdResolution | None = None,
) -> CommandModuleMap:
    if resolution:
        return resolution(
            current_command_item.subcommands,
            new_command_item.subcommands,
        )
    return {
        **current_command_item.subcommands,
        **new_command_item.subcommands,
    }



from typing import Any

from m.core import Bad, Good, Res, issue, one_of
from m.log import EnvVars
from pydantic import BaseModel

from .config import Config, Workflow
from .git_env import GitEnv


class ReleaseEnv(BaseModel):
    """Object to store the release configuration."""

    build_tag: str
    python_tag: str
    is_release: bool
    is_release_pr: bool
    is_hotfix_pr: bool
    workflow: Workflow


def _verify_version(
    config: Config,
    git_env: GitEnv,
    gh_latest: str,
    is_release_pr: bool,
    is_release: bool,
) -> Res[int]:
    if config.workflow in {Workflow.git_flow, Workflow.m_flow}:
        if config.uses_git_flow():
            pr_branch = git_env.get_pr_branch()
            flow = config.git_flow
            prefixes = (flow.release_prefix, flow.hotfix_prefix)
            # Skip verification when release or hotfix are going to develop
            if git_env.target_branch == flow.develop_branch:
                if pr_branch.startswith(prefixes):
                    return Good(0)
        return config.verify_version(
            gh_latest,
            is_release_pr=is_release_pr,
            is_release=is_release,
        )
    # Covers Workflow.free_flow
    return Good(0)


def _get_develop_branch(config: Config) -> str:
    if config.uses_git_flow():
        return config.git_flow.develop_branch
    return 'develop'


def _extra_checks(
    config: Config,
    git_env: GitEnv,
    is_release_pr: bool,
    is_hotfix_pr: bool,
) -> Res[Any]:
    # If successful we return None, we do not care about the value thus we
    # are specifying `Any` so that a Bad value may be compatible with other
    # `OneOf`s.
    master_branch = config.get_master_branch()
    develop_branch = _get_develop_branch(config)
    valid_branches = (
        (master_branch,)
        if config.uses_m_flow()
        else (master_branch, develop_branch)
    )
    release_pr = is_release_pr or is_hotfix_pr
    if release_pr and git_env.target_branch not in valid_branches:
        error_type = 'release' if is_release_pr else 'hotfix'
        return issue(f'invalid {error_type}-pr', context={
            'expected_target_branch': master_branch,
            'current_target_branch': git_env.target_branch,
            'workflow': str(config.workflow),
        })
    return Good(None)


def get_release_env(
    config: Config,
    env_vars: EnvVars,
    git_env: GitEnv,
) -> Res[ReleaseEnv]:
    """Provide the release environment information.

    Args:
        config: The m configuration.
        env_vars: The environment variables.
        git_env: The git environment.

    Returns:
        A `ReleaseEnv` instance.
    """
    is_release = git_env.is_release(config)
    is_release_pr = git_env.is_release_pr(config)
    is_hotfix_pr = git_env.is_hotfix_pr(config)
    gh_latest = git_env.release.tag_name if git_env.release else ''
    if not config.uses_free_flow():
        check_result = _extra_checks(
            config,
            git_env,
            is_release_pr=is_release_pr,
            is_hotfix_pr=is_hotfix_pr,
        )
        if isinstance(check_result, Bad):
            return check_result
    return one_of(lambda: [
        ReleaseEnv(
            build_tag=build_tag,
            python_tag=python_tag,
            is_release=is_release,
            is_release_pr=is_release_pr,
            is_hotfix_pr=is_hotfix_pr,
            workflow=config.workflow,
        )
        for _ in _verify_version(
            config,
            git_env,
            gh_latest,
            is_release_pr=is_release_pr or is_hotfix_pr,
            is_release=is_release,
        )
        for build_tag in git_env.get_build_tag(config, env_vars.run_id)
        for python_tag in git_env.get_py_tag(config, env_vars.run_id)
    ])



import os
from functools import partial
from pathlib import Path
from types import MappingProxyType

from m.color.colors import color
from m.core import Bad, subprocess

from m import git

UNKNOWN = 'unknown'
UNKNOWN_COLOR = '\033[38;5;20m'

COL = partial(color, auto_end=False)

# https://www.compart.com/en/unicode/
# https://gitmoji.dev/
STATUS_SYMBOLS = MappingProxyType({
    UNKNOWN: '🧐',  # contender: ???
    'untracked': '🌱',
    'stash': '\u271A',  # Heavy Greek Cross
    'ahead': '\u25B6',  # Black Right-Pointing Triangle
    'behind': '\u25C0',  # Black Left-Pointing Triangle
    'clean': '\u2714',  # Heavy Check Mark
    'staged': '🌭',  # contender: \u25CF Black Circle
    'dirty': '\u2716',  # Heavy Multiplication X
    'diverged': '💥',  # contender: !
})

STATUS_COLORS = MappingProxyType({
    UNKNOWN: UNKNOWN_COLOR,  # dark blue
    'untracked': '\033[38;5;76m',  # mid lime-green
    'stash': '\033[38;5;76m',  # mid lime-green
    'ahead': '\033[38;5;226m',  # bright yellow
    'behind': '\033[38;5;142m',  # darker yellow-orange
    'clean': '\033[38;5;82m',  # brighter green
    'staged': '\033[38;5;214m',  # orangey yellow
    'dirty': '\033[38;5;202m',  # orange
    'diverged': '\033[38;5;196m',  # red
})


def _branch_display(current_branch: str) -> str:
    # if the branch is HEAD, display the commit hash instead
    if current_branch == 'HEAD':
        short_res = subprocess.eval_cmd('git rev-parse --short HEAD')
        return short_res.get_or_else(current_branch)
    return current_branch


def _branch_sec(branch: str, status: str) -> str:
    s_color = STATUS_COLORS.get(status, UNKNOWN_COLOR)
    s_sym = STATUS_SYMBOLS.get(status, '???')
    branch_info = ''.join([r'\[', s_color, r'\]', f'{s_sym} {branch}'])
    return COL(r'\[{gray}\][', branch_info, r'\[{gray}\]]')


def _repo_info() -> tuple[str, str]:
    repo_path = git.get_repo_path().get_or_else('[UNKNOWN_REPO_PATH]')
    cwd = str(Path.cwd())
    repo = repo_path.split('/')[-1]
    rel_path = cwd.replace(repo_path, '^', 1)
    return repo, rel_path


def _container_info() -> str:
    name = os.environ.get('DK_CONTAINER_NAME', 'devcontainer')
    version = os.environ.get('DK_CONTAINER_VERSION')
    if version:
        dk_version = version.replace('0.0.0-', '')
        if dk_version.startswith('local'):
            dk_version = '[DEV]'
        return f'{name}@{dk_version}'
    return name


def _git_prompter(current_branch: str) -> str:
    branch = _branch_display(current_branch)
    status_res = git.get_status(check_stash=True)
    status, _ = status_res.get_or_else((UNKNOWN, UNKNOWN))
    status_color = STATUS_COLORS.get(status, UNKNOWN_COLOR)
    arrow = ''.join([r'\[', status_color, r'\]', '\u279C'])
    container_sec = COL(r'\[{green}\]', _container_info(), r'\[{white}\]:')
    branch_sec = _branch_sec(branch, status)
    repo, rel_path = _repo_info()
    repo_sec = COL(r'\[{blue}\]', repo)
    relpath_sec = COL(r'\[{gray}\]', rel_path)
    end = COL(r'\[{white}\]$')
    return f'{arrow} {container_sec}{repo_sec} {branch_sec} {relpath_sec}{end} '


def prompter() -> str:
    """Command line prompter.

    Returns:
        The string to be displayed in the prompt.
    """
    branch_res = git.get_branch()
    if isinstance(branch_res, Bad):
        return COL(r'\[{orange}\]\w\[{end}\]$ ')
    return _git_prompter(branch_res.value)



import argparse
import sys
from os import path as pth
from typing import Any, Callable, cast
from warnings import warn

import argcomplete
import typing_extensions
from m.core import Bad, Issue, OneOf
from m.log import Logger

from ..core.io import env
from .engine.misc import params_count
from .engine.sys import default_meta, default_root_meta, import_cli_commands
from .engine.types import (
    CliCommands,
    CliSubcommands,
    CommandModule,
    MetaModule,
)
from .handlers import create_issue_handler, create_json_handler
from .validators import validate_non_empty_str

logger = Logger('m.cli')
cli_global_option = Callable[[argparse.ArgumentParser], argparse.Action]


default_result_handler = create_json_handler(pretty=False)
default_issue_handler = create_issue_handler(use_warning=False)


def _main_parser(cli_commands: CliCommands) -> argparse.Namespace:
    root_meta_module = cli_commands.meta or default_root_meta
    root_meta = root_meta_module.meta
    raw = argparse.RawTextHelpFormatter
    # NOTE: In the future we will need to extend from this class to be able to
    #   override the error method to be able to print CI environment messages.
    argp = argparse.ArgumentParser(
        formatter_class=raw,
        description=root_meta['description'] if root_meta else '...',
    )
    if root_meta_module.add_arguments:
        root_meta_module.add_arguments(argp)
    subp = argp.add_subparsers(
        title='commands',
        dest='command_name',
        required=True,
        help='additional help',
        metavar='<command>',
    )
    names = sorted(cli_commands.commands.keys())
    for name in names:
        cmd_item = cli_commands.commands[name]
        if isinstance(cmd_item, CliSubcommands):
            subcmd_mod = cmd_item
            meta_mod = subcmd_mod.meta or default_meta
            meta_dict = meta_mod.meta
            parser = subp.add_parser(
                name,
                help=meta_dict['help'],
                formatter_class=raw,
                description=meta_dict['description'],
            )
            if meta_mod.add_arguments:
                meta_mod.add_arguments(parser)
            subsubp = parser.add_subparsers(
                title='commands',
                dest='subcommand_name',
                required=True,
                help='additional help',
                metavar='<command>',
            )
            sub_mod = subcmd_mod.subcommands
            for subname in sorted(sub_mod.keys()):
                run_func = sub_mod[subname].run
                if params_count(run_func) == 3:
                    run_func(subname, None, subsubp)
        else:
            mod_inst = cmd_item
            run_func = mod_inst.run
            if params_count(run_func) == 3:
                run_func(name, None, subp)
    argcomplete.autocomplete(argp)
    return argp.parse_args()


def exec_cli(cli_commands: CliCommands) -> None:
    """Execute the cli application.

    usage::

        def create_cli_commands() -> CliCommands:
            # We may import CliCommand objects from other projects and create
            # a new one with them.
            return import_cli_commands('cli.command.module')

        def main():
            cli_commands = create_cli_commands()
            exec_cli(cli_commands)

    This is the preferred way to execute the cli application as it will allow
    other potential applications to use the cli commands.

    Args:
        cli_commands: The cli commands to execute.
    """
    arg = _main_parser(cli_commands)

    run_func = None
    command_name = ''

    commands = cli_commands.commands
    # WPS421 encourages to use try/except instead of hasattr but in this
    # case we want explicitly before using it.
    if hasattr(arg, 'subcommand_name'):  # noqa: WPS421
        command_name = arg.subcommand_name
        sub_mod = cast(CliSubcommands, commands[arg.command_name])
        run_func = sub_mod.subcommands[command_name].run
    else:
        command_name = arg.command_name
        run_func = cast(CommandModule, commands[command_name]).run

    len_run_args = params_count(run_func)
    run_args = [command_name, arg, None][:len_run_args]
    exit_code = 0
    try:
        # mypy is having a hard time figuring out the type of run_args
        exit_code = run_func(*run_args)  # type:ignore[arg-type]
    except Exception as ex:
        default_issue_handler(
            Issue('unknown cli run function exception', cause=ex),
        )
        exit_code = 5
    sys.exit(exit_code)


@typing_extensions.deprecated(
    '`run_cli` deprecated; use `exec_cli` instead.',
)
def run_cli(
    commands_module: str | None,
    add_args: Callable[[argparse.ArgumentParser], None] | None = None,
) -> None:  # pragma: no cover
    """Run the cli application.

    Deprecated, use `exec_cli` instead.

    usage::

        def add_args(argp):
            argp.add_argument(...)
        def main():
            run_cli('m.cli.commands', add_args)

    We only need `add_args` if we need to gain access to the
    `argparse.ArgumentParser` instance.

    Args:
        commands_module: The full name of the module containing the commands.
        add_args: Optional callback to gain access to the ArgumentParser.
    """
    # NOTE: This is a deprecated feature and will be removed in the future.
    if commands_module and '/' in commands_module:
        warn(
            '`run_cli(__file__)` is deprecated, use `run_cli("commands.module") instead',
            DeprecationWarning,
        )
        root = pth.split(pth.split(commands_module)[0])[1]
        commands_module = f'{root}.cli.commands'
    cli_commands: CliCommands = CliCommands(commands={}, meta=default_root_meta)
    if commands_module:
        cli_commands = import_cli_commands(commands_module)
    if add_args:
        warn('run_cli add_args is deprecated, use meta.add_arguments instead', DeprecationWarning)
        if not cli_commands.meta:
            cli_commands.meta = MetaModule(
                meta=default_root_meta.meta,
                add_arguments=add_args,
            )
        cli_commands.meta.add_arguments = add_args
    exec_cli(cli_commands)


def run_main(
    callback: Callable[[], OneOf[Issue, Any]],
    result_handler: Callable[[Any], None] = default_result_handler,
    issue_handler: Callable[[Issue], None] = default_issue_handler,
) -> int:
    """Run the callback and print the returned value.

    To change how the result or an issue should be display provide the optional
    arguments `handle_result` and `handle_issue`. For instance, to display the
    raw value provide the `print` function.

    Args:
        callback: A function that returns a `OneOf`.
        result_handler: A function that takes in the Good result.
        issue_handler: A function that takes in the Issue.

    Returns:
        0 if the callback is a `Good` result otherwise return 1.
    """
    res = None
    try:
        res = callback()
    except Exception as ex:
        issue_handler(Issue('unknown caught exception', cause=ex))
        return 2
    if isinstance(res, Bad):
        problem = res.value
        if isinstance(problem, Issue):
            issue_handler(problem)
        else:
            issue_handler(Issue('non-issue exception', cause=problem))
        return 1
    result_handler(res.value)
    return 0


def cli_integration_token(
    integration: str,
    env_var: str,
) -> cli_global_option:
    """Return a function that takes in a parser.

    This generated function registers a token argument in the parser
    which looks for its value in the environment variables.

    Args:
        integration: The name of the integration.
        env_var: The environment variable name.

    Returns:
        A function to add arguments to an argparse parser.
    """
    return lambda parser: parser.add_argument(
        '-t',
        '--token',
        type=validate_non_empty_str,
        default=env(env_var),
        help=f'{integration} access token (default: env.{env_var})',
    )



import os
import sys
from collections.abc import Callable
from inspect import signature
from typing import Any, TypeVar

from m.cli import run_main
from m.cli.cli import default_issue_handler
from m.core import Bad, Good, Issue, Res, issue, rw
from m.log import logging_config
from m.pydantic import KebabModel
from pydantic import TypeAdapter
from pydantic.fields import FieldInfo

InputModel = TypeVar('InputModel', bound=KebabModel)
OutputModel = TypeVar('OutputModel', bound=KebabModel)


def load_step_inputs(model: type[InputModel]) -> Res[InputModel]:
    """Load the GitHub Action inputs from the environment.

    This function can be used provided that in the GitHub Action workflow we
    set up:::

        env:
            INPUT_ARG_1: ${{ inputs.arg-1 }}

    In this way the inputs will set in the environment which we can then read
    in our main python function.

    Args:
        model: The class to create an instance of.

    Returns:
        A `OneOf` with the model or an issue.
    """
    model_data: dict[str, str] = {}
    for key, env_val in os.environ.items():
        if key.startswith('INPUT_'):
            _, input_name = key.split('INPUT_')
            model_data[input_name.lower()] = env_val
    try:
        return Good(TypeAdapter(model).validate_python(model_data))
    except Exception as ex:
        return issue('load_step_inputs_failure', cause=ex)


def dump_step_outputs(outputs: OutputModel) -> Res[int]:
    """Dump the outputs for a GitHub Action step.

    Github does this if we write to the file where $GITHUB_OUTPUT is pointing.

    Args:
        outputs: The outputs to dump.

    Returns:
        A `OneOf` with 0 or an issue.
    """
    output_data = '\n'.join([
        f'{key}={output_val}'
        for key, output_val in outputs.model_dump(by_alias=True).items()
    ])
    output_file = os.environ.get('GITHUB_OUTPUT', 'NO_GITHUB_OUTPUT_FILE')
    return rw.write_file(output_file, output_data, 'a')


def _result_handler(outputs: OutputModel) -> None:
    try:
        res = dump_step_outputs(outputs)
    except Exception as ex:
        default_issue_handler(Issue('dump_step_outputs', cause=ex))
        sys.exit(5)
    if isinstance(res, Bad):
        default_issue_handler(res.value)
        sys.exit(6)


def run_action(main: Callable[[InputModel], Res[OutputModel]]) -> None:
    """Entry point for a GitHub Action.

    This is the main function that should be used to run an action. It takes in
    a function that takes in a [m.pydantic.KebabModel][] and returns a
    `Res[KebabModel]`.

    The only place where this function is needed is in the the if block

    ```python
    if __name__ == '__main__':
        run_action(my_action)
    ```

    `mypy` will make sure that the you are providing the correct type of function
    to `run_action`. Keep in mind, the function is generic and we should be writing
    models for the inputs and outputs for all of our functions.

    Args:
        main: The main function of the GitHub Action.
    """
    logging_config()
    main_params = signature(main).parameters
    arg_name = next(iter(main_params))
    input_model = main_params[arg_name].annotation
    args = load_step_inputs(input_model)
    if isinstance(args, Bad):
        default_issue_handler(args.value)
        sys.exit(4)
    inputs = args.value
    exit_code = run_main(lambda: main(inputs), result_handler=_result_handler)
    sys.exit(exit_code)


def InArg(  # noqa: N802
    *,
    help: str,  # noqa: WPS125
    default: str | None = None,
) -> Any:
    """Force proper annotation of the input of a GitHub Action.

    Should be used to declare the input arguments of an action. It returns
    [`Any`][typing.Any] to bypass `mypy`'s type checking. Similar to
    [pydantic.fields.Field][] but it is tailored to help us write the inputs for
    an action and its steps.

    !!! note

        By default all input arguments are required. If you want to make an
        input not required then provide a default value.

    Args:
        help: Human-readable description.
        default: The default value for the argument.

    Returns:
        A new [`FieldInfo`][pydantic.fields.FieldInfo], the return annotation is
            `Any` so `InArg` can be used on type annotated fields without
            causing typing errors.
    """
    args = {
        'description': help,
    }
    if default is not None:
        args['default'] = default
    return FieldInfo.from_field(**args)


def OutArg(  # noqa: N802
    *,
    help: str,  # noqa: WPS125
    export: bool = False,
) -> Any:
    """Force proper annotation of the output of a GitHub Action.

    !!! note
        All steps have access to the steps output, if we want to make the output
        available to the action we need to `export` it.

    Args:
        help: Human-readable description.
        export: Whether the argument is to be exported to the action.

    Returns:
        A new [`FieldInfo`][pydantic.fields.FieldInfo], the return annotation is
            `Any` so `Arg` can be used on type annotated fields without causing
            a typing error.
    """
    return FieldInfo.from_field(
        description=help,
        json_schema_extra={'export': export},
    )



from m.cli import (
    Arg,
    BaseModel,
    command,
    create_json_handler,
    create_yaml_handler,
    run_main,
)


class Arguments(BaseModel):
    """Create the [m_dir]/.m/env.list file.

    With the `bashrc` option it will print a bashrc snippet and no file will
    be created. Note that boolean values are lowercased in the snippet. This
    eventually be the case for generated env.list file.
    """

    pretty: bool = Arg(
        default=False,
        help='format json payload with indentation',
    )
    yaml: bool = Arg(
        default=False,
        help='use yaml format',
    )
    m_dir: str = Arg(
        default='m',
        help='m project directory',
        positional=True,
    )
    bashrc: bool = Arg(
        default=False,
        help='print bashrc snippet',
    )


@command(
    help='create a list of env variables',
    model=Arguments,
)
def run(arg: Arguments) -> int:
    from m.ci.m_env import bashrc_snippet, write_m_env_vars

    if arg.bashrc:
        return run_main(
            lambda: bashrc_snippet(arg.m_dir),
            result_handler=print,
        )

    result_handler = (
        create_yaml_handler(arg.pretty)
        if arg.yaml
        else create_json_handler(arg.pretty)
    )
    return run_main(
        lambda: write_m_env_vars(arg.m_dir),
        result_handler=result_handler,
    )



import re
from datetime import datetime

from m.ci.config import Config, get_m_filename, read_config
from m.core import Good, Res, issue, one_of, rw
from m.git import get_first_commit_sha
from m.github.ci import compare_sha_url
from m.log import Logger

logger = Logger('m.ci.release_setup')


def _get_versions(lines: list[str], new_ver: str, first_sha: str) -> list[str]:
    versions = [new_ver]
    for line in lines:
        match = re.match(r'## \[(.*)]', line)
        if match:
            versions.append(match.groups()[0])
    versions.append(first_sha)
    return versions


def _version_anchor(ver: str) -> str:
    return f'<a name="{ver}" href="#{ver}">-</a>'


def new_changelog(
    changelog_contents: str,
    owner: str,
    repo: str,
    new_ver: str,
    first_sha: str,
) -> Res[str]:
    """Modify the contents of a CHANGELOG.

    It adds a new entry with the new version the new changelog contents.

    Args:
        changelog_contents: The current contents of the CHANGELOG file.
        owner: The repo owner.
        repo: The repo name.
        new_ver: The new version.
        first_sha: The very first commit sha of the repo.

    Returns:
        The new contents of the CHANGELOG.
    """
    parts = changelog_contents.split('## [Unreleased]')
    if len(parts) != 2:
        return issue('missing "Unreleased" link')

    header, main = parts
    entries = main.split('[unreleased]:')[0]
    versions = _get_versions(entries.split('\n'), new_ver, first_sha)

    compare_url = compare_sha_url(owner, repo, new_ver, 'HEAD')
    links = [f'[unreleased]: {compare_url}']
    for i in range(len(versions) - 1):
        link = compare_sha_url(owner, repo, versions[i + 1], versions[i])
        links.append(f'[{versions[i]}]: {link}')

    date = datetime.now().strftime('%B %d, %Y')
    ver_anchor = _version_anchor(new_ver)
    return Good(''.join([
        header,
        '## [Unreleased]\n\n',
        f'## [{new_ver}] {ver_anchor} {date}\n\n',
        entries,
        '\n'.join(links),
        '\n',
    ]))


def update_changelog_file(
    owner: str,
    repo: str,
    new_ver: str,
    first_sha: str,
    filename: str = 'CHANGELOG.md',
) -> Res[int]:
    """Add the new version entry to be released to the CHANGELOG.

    Args:
        owner: The repo owner.
        repo: The repo name.
        new_ver: The version that is being released.
        first_sha: The first sha ever committed on the repo.
        filename: Specify the CHANGELOG file (defaults to CHANGELOG.md)

    Returns:
        0 if successful, an issue otherwise.
    """
    return one_of(lambda: [
        0
        for text in rw.read_file(filename)
        for new_data in new_changelog(text, owner, repo, new_ver, first_sha)
        for _ in rw.write_file(filename, new_data)
        for _ in logger.info(f'updated {filename}')
    ])


def _update_line_version(line: str, ver: str) -> str:
    stripped_line = line.strip()
    if stripped_line.startswith('"version"'):
        comma = ',' if stripped_line.endswith(',') else ''
        return f'  "version": "{ver}"{comma}'
    if stripped_line.startswith('version'):
        return f'version: {ver}'
    return line


def _update_config_version(
    config_contents: str,
    ver: str,
) -> Res[str]:
    lines = config_contents.split('\n')
    new_lines = [_update_line_version(line, ver) for line in lines]
    return Good('\n'.join(new_lines))


def update_version(
    root: str,
    version: str,
) -> Res[int]:
    """Update the version property in m configuration file.

    Args:
        root: The directory with the m configuration file.
        version: The new version to write in the m configuration.

    Returns:
        0 if successful or an issue.
    """
    return one_of(lambda: [
        0
        for filename in get_m_filename(root)
        for config_contents in rw.read_file(filename)
        for new_data in _update_config_version(config_contents, version)
        for _ in rw.write_file(filename, new_data)
        for _ in logger.info(f'bumped version in {filename}')
    ])


def _success_release_setup(config: Config, new_ver: str) -> Res[int]:
    link = compare_sha_url(config.owner, config.repo, config.version, 'HEAD')
    return logger.info('setup complete', {
        'new_version': new_ver,
        'unreleased_changes': link,
    })


def _read_config(m_dir: str, config: Config | None) -> Res[Config]:
    if config:
        return Good(config)
    return read_config(m_dir)


def release_setup(
    m_dir: str,
    config_inst: Config | None,
    new_ver: str,
    changelog: str = 'CHANGELOG.md',
) -> Res[None]:
    """Modify all the necessary files to create a release.

    These include: CHANGELOG.md and the m configuration file.

    Args:
        m_dir: The directory with the m configuration.
        config_inst: If provided it skips reading the configuration.
        new_ver: The new version to write in the m configuration.
        changelog: The name of the changelog file (defaults to CHANGELOG.md)

    Returns:
        None if successful, otherwise an issue.
    """
    return one_of(lambda: [
        None
        for config in _read_config(m_dir, config_inst)
        for first_sha in get_first_commit_sha()
        for _ in update_version(m_dir, new_ver)
        for _ in update_changelog_file(
            config.owner,
            config.repo,
            new_ver,
            first_sha,
            changelog,
        )
        for _ in _success_release_setup(config, new_ver)
    ])



import sys
from typing import Any

from m.core import Bad, Good, Res, issue, one_of, rw
from m.pydantic import KebabModel

from .action_file import ActionFile
from .actions import Action
from .system import import_attr


def build_actions(
    actions_py_file: str,
    *,
    check: bool,
    show_traceback: bool,
) -> Res[None]:
    """Create the actions metadata files as described by the actions file.

    Args:
        actions_py_file: The path to a python script containing the actions.
        check: Check if generated files are up to date and skip writing them.
        show_traceback: Display traceback in case of an issue.

    Returns:
        A `Good` containing `None` if the files were written.
    """
    return one_of(lambda: [
        None
        for _ in rw.assert_file_exists(actions_py_file)
        for py_path, actions_module in update_py_path(actions_py_file)
        for possible_actions in import_attr(f'{actions_module}.actions')
        for actions in assert_actions(possible_actions)
        for _ in process_actions(
            py_path,
            actions_module,
            actions,
            check=check,
            show_traceback=show_traceback,
        )
    ])


def update_py_path(actions_py_file: str) -> Res[tuple[str, str]]:
    """Obtain the python path and module name from the actions file.

    It also modifies the PYTHONPATH so that other files may be imported from
    the same directory.

    Args:
        actions_py_file: The path to a python script containing the actions.

    Returns:
        A `Good` containing a tuple of the python path and module name.
    """
    is_file = actions_py_file.endswith('.py')
    parts = actions_py_file.split('/')
    # Need to add the parent directory to the path so that we can do relative
    # imports from the actions file.
    py_path = (
        '/'.join(parts[:-2])
        if is_file
        else '/'.join(parts[:-1])
    )
    sys.path.insert(0, py_path)
    module_name = parts[-1].split('.')[0]
    dir_name = parts[-2] if is_file else ''
    module = f'{dir_name}.{module_name}' if is_file else module_name
    return Good((py_path, module))


def assert_actions(actions: Any) -> Res[list[Action]]:
    """Assert that the actions instance is of type `Action` or `list[Action]`.

    Args:
        actions: The actions instance to assert.

    Returns:
        A `Good` containing a list of `Action`.
    """
    if isinstance(actions, Action):
        return Good([actions])
    if isinstance(actions, list) and all(isinstance(a, Action) for a in actions):
        return Good(actions)
    suggestion = 'ensure all actions are of type "m.github.actions.Action"'
    return issue('invalid_actions', context={'suggestion': suggestion})


def process_actions(
    py_path: str,
    mod_name: str,
    actions: list[Action],
    *,
    check: bool,
    show_traceback: bool,
) -> Res[None]:
    """Iterate over each action to either write or check the action file.

    Args:
        py_path: The root directory of the actions.
        mod_name: The name of the module containing the actions.
        actions: The list of actions to process.
        check: If True, it verifies the action file is up to date.
        show_traceback: Display traceback in case of an issue.

    Returns:
        An `Good` containing None or an issue.
    """
    process_result = [
        process_action(py_path, mod_name, action, check=check)
        for action in actions
    ]
    issues = {
        action.name: action_res.value
        for action_res, action in zip(process_result, actions)
        if isinstance(action_res, Bad)
    }
    if issues:
        return issue(
            'process_actions_failure',
            context={
                'issues': {
                    action_name: iss.to_dict(show_traceback=show_traceback)
                    for action_name, iss in issues.items()
                },
            },
        )
    return Good(None)


def process_action(
    py_path: str,
    actions_module: str,
    action: Action,
    *,
    check: bool,
) -> Res[int]:
    """Process an action to either write or check the action file.

    Args:
        py_path: The root directory of the actions.
        actions_module: The name of the module containing the actions.
        action: The action to process.
        check: If True, it verifies the action file is up to date.

    Returns:
        An `Good` containing None or an issue.
    """
    outputs_res = action.gather_outputs()
    if isinstance(outputs_res, Bad):
        return Bad(outputs_res.value)

    outputs, available_outputs = outputs_res.value
    action_file = ActionFile(
        py_path=py_path,
        module_name=actions_module,
        name=action.name,
        description=action.description,
        inputs=action.inputs or KebabModel,
        action=action,
        outputs=outputs,
        python_path=py_path,
        available_outputs=available_outputs,
    )
    yaml_contents = str(action_file)
    if check:
        current_contents_res = rw.read_file(action.file_path)
        if isinstance(current_contents_res, Bad):
            return Bad(current_contents_res.value)
        current_contents = current_contents_res.value
        if current_contents != yaml_contents:
            return issue(
                'action_file_out_of_date',
                context={
                    'action_name': action.name,
                    'file_path': action.file_path,
                },
            )
        return Good(0)
    return rw.write_file(action.file_path, yaml_contents)



import os
import sys
from argparse import ArgumentTypeError
from pathlib import Path
from typing import Any

from ..core.json import parse_json, read_json
from ..core.yaml_fp import read_yson


def validate_json_payload(file_path: str) -> Any:
    """Return a dictionary from the contents of file_path.

    This is a string that tell us to read from a file, stdin or just
    plain json data.

    It can be used parse `yaml` files as well. The extension should be
    `.yaml` or `.yml`.

    Args:
        file_path: A string with the json payload. If it starts with `@` then
            the name of a valid json file from where the payload will be read.

    Raises:
        ArgumentTypeError: If the file_path is meant to be a valid path and it
            does not exist.

    Returns:
        A parsed json payload
    """
    if file_path == '@-':
        res = read_json(None)
        if res.is_bad:
            msg = f'invalid json payload in SYS.STDIN\n{res.value}'
            raise ArgumentTypeError(msg)
        return res.value
    if file_path.startswith('@'):
        err = ''
        filename = file_path[1:]
        if Path.exists(Path(filename)):
            res = read_yson(filename)
            if not res.is_bad:
                return res.value
            err = f'invalid json payload in {filename}\n{res.value}'
        else:
            err = f'file "{filename}" does not exist'
        if err:
            raise ArgumentTypeError(err)
    res = parse_json(file_path)
    if res.is_bad:
        raise ArgumentTypeError(f'invalid json payload\n{res.value}')
    return res.value


def validate_payload(file_path: str) -> str:
    """Return the raw payload.

    This allows us to read from a file or the stdin stream.

    Args:
        file_path: A string with the payload. If it starts with `@` then the
            name of a valid file from where the payload will be read.

    Raises:
        ArgumentTypeError: If the file_path is meant to be a valid path and it
            does not exist.

    Returns:
        The payload found in the file.
    """
    if file_path.startswith(r'\@'):
        # escape @ with \ to let the cli know that the payload starts with @
        return file_path[1:]
    if file_path == '@-':
        return sys.stdin.read()
    if file_path.startswith('@'):
        filename = file_path[1:]
        path_handle = Path(filename)
        if not Path.exists(path_handle):
            raise ArgumentTypeError(f'file "{filename}" does not exist')
        with Path.open(path_handle, encoding='UTF-8') as fp:
            return fp.read()
    return file_path


def validate_non_empty_str(arg_value: str) -> str:
    """Return the value as long as its not empty.

    Args:
        arg_value: The input provided by argparse.

    Raises:
        ArgumentTypeError: If the input is not provided.

    Returns:
        The value if non empty.
    """
    if not arg_value:
        raise ArgumentTypeError('empty value not allowed')
    return arg_value


def env_var(arg_value: str) -> str:
    """Read a value from the environment.

    Args:
        arg_value: The input provided by argparse.

    Returns:
        The environment variable value if it exists. Otherwise, the input.
    """
    return os.environ.get(arg_value, arg_value)


def env_var_or_empty(arg_value: str) -> str:
    """Read a value from the environment.

    Unlike the [env_var][m.cli.validators.env_var] validator, this will only
    allow the `arg_value` to pass through if is not in the form of an
    environment variable. That is, if the value is all uppercase letters and
    underscores it will attempt to read from the environment and return an empty
    string if not defined.

    Args:
        arg_value: The input provided by argparse.

    Returns:
        The environment variable value if it exists. Otherwise, empty string.
    """
    env_form = arg_value.upper().replace('-', '_', -1)
    if arg_value == env_form:
        return os.environ.get(arg_value, '')
    return arg_value


def validate_file_exists(path: str) -> str:
    """Assert that a file exists.

    Args:
        path: The path to the file.

    Raises:
        ArgumentTypeError: If the file does not exist.

    Returns:
        The path if it exists.
    """
    path_inst = Path(path)
    if path_inst.exists():
        return path
    raise ArgumentTypeError('file does not exist')



from typing import List

from ....core import Issue, OneOf, json, one_of
from ..core.types import FileReport, Violation


def read_payload(payload: str) -> OneOf[Issue, List[FileReport]]:
    """Transform an eslint payload to a list of `FileReport` instances.

    Args:
        payload: The raw payload from eslint.

    Returns:
        A `OneOf` containing an `Issue` or a list of `FileReport` instances.
    """
    return one_of(lambda: [
        [
            FileReport(
                file_path=error['filePath'],
                violations=[
                    Violation(
                        msg['ruleId'],
                        msg['message'],
                        msg['line'],
                        msg['column'],
                        error['filePath'],
                    )
                    for msg in error['messages']
                ],
            )
            for error in json_payload
        ]
        for json_payload in json.parse_json(payload)
    ])



from typing import List, Optional

from m.pydantic import CamelModel
from pydantic import BaseModel


class Author(CamelModel):
    """An object representing a committer."""

    login: str
    avatar_url: str | None = None
    email: str | None = None


class AssociatedPullRequest(BaseModel):
    """Information for commits that are associated with a pull request."""

    # pylint: disable=too-many-instance-attributes
    author: Author
    merged: bool
    pr_number: int
    target_branch: str
    target_sha: str
    pr_branch: str
    pr_sha: str
    title: str
    body: str


class Commit(BaseModel):
    """The git commit info."""

    author_login: str
    short_sha: str
    sha: str
    message: str
    url: str
    associated_pull_request: Optional[AssociatedPullRequest] = None

    def get_pr_branch(self) -> str:
        """Find the git branch if commit has an associated pr.

        Returns:
            The pr branch if the commit has an associated pr or empty
            string.
        """
        if not self.associated_pull_request:
            return ''
        return self.associated_pull_request.pr_branch

    def is_release(self, release_prefix: str | None) -> bool:
        """Determine if the current commit should create a release.

        Args:
            release_prefix: The prefix used to flag a release.

        Returns:
            True if we are dealing with a release branch.
        """
        if not release_prefix:
            return False
        return self.get_pr_branch().startswith(f'{release_prefix}/')


class PullRequest(BaseModel):
    """Pull request information."""

    # pylint: disable=too-many-instance-attributes
    author: Author
    pr_number: int
    pr_branch: str
    target_branch: str
    target_sha: str
    url: str
    title: str
    body: str
    file_count: int
    files: List[str]
    is_draft: bool

    def is_release_pr(self, release_prefix: str | None) -> bool:
        """Determine if the pull request is a release pull request.

        Args:
            release_prefix: The prefix used to flag a release.

        Returns:
            True if we are dealing with a release branch.
        """
        if not release_prefix:
            return False
        return self.pr_branch.startswith(f'{release_prefix}/')


class Release(BaseModel):
    """Github release <==> Git tag."""

    name: str
    tag_name: str
    published_at: str


class GithubCiRunInfo(BaseModel):
    """The main information we need for a ci run."""

    commit: Commit
    pull_request: Optional[PullRequest] = None
    release: Optional[Release] = None


class CommitInfo(BaseModel):
    """A commit can be tracked with the following properties."""

    owner: str
    repo: str
    sha: str



from pathlib import Path

from packaging.version import Version
from pydantic import BaseModel, ConfigDict

from ..core import Good, Res, hone, issue, one_of, yaml_fp
from .docker.config import DockerConfig
from .types import Branches, GitFlowConfig, MFlowConfig, Workflow


def _handle_release_pr(ver_gt_latest: bool) -> str:
    return '' if ver_gt_latest else 'version needs to be bumped'


def _handle_release(ver_gt_latest: bool) -> str:
    return '' if ver_gt_latest else 'version was not bumped during release pr'


def _handle_non_release(ver_lt_latest: bool, ver_gt_latest: bool) -> str:
    if ver_lt_latest:
        return 'version is behind (Branch may need to be updated)'
    if ver_gt_latest:
        return 'version is ahead (Revert configuration change)'
    return ''


class Config(BaseModel):
    """Object to store the m project configuration."""

    # pylint: disable=too-many-instance-attributes
    m_dir: str
    owner: str
    repo: str
    version: str = '0.0.0'
    workflow: Workflow = Workflow.free_flow
    git_flow: GitFlowConfig = GitFlowConfig(
        master_branch=Branches.master,
        develop_branch=Branches.develop,
        release_prefix=Branches.release,
        hotfix_prefix=Branches.hotfix,
    )
    m_flow: MFlowConfig = MFlowConfig(
        master_branch=Branches.master,
        release_prefix=Branches.release,
        hotfix_prefix=Branches.hotfix,
    )
    build_tag_with_version: bool = False
    docker_config: DockerConfig | None = None

    model_config = ConfigDict(use_enum_values=True)

    def uses_git_flow(self) -> bool:
        """Check if configuration is using the git flow.

        Returns:
            True if workflow is the git flow.
        """
        return self.workflow == Workflow.git_flow

    def uses_m_flow(self) -> bool:
        """Check if configuration is using the m flow.

        Returns:
            True if workflow is the m flow.
        """
        return self.workflow == Workflow.m_flow

    def uses_free_flow(self) -> bool:
        """Check if configuration is using the free flow.

        Returns:
            True if workflow is the free flow.
        """
        return self.workflow == Workflow.free_flow

    def get_master_branch(self) -> str:
        """Obtain the name of the branch that aliases the "master" branch.

        Returns:
            The name/alias assigned to the `master` branch.
        """
        if self.uses_m_flow():
            return self.m_flow.master_branch
        if self.uses_git_flow():
            return self.git_flow.master_branch
        return 'master'

    def get_develop_branch(self) -> str:
        """Obtain the name of the branch that aliases the "develop" branch.

        Returns:
            The name/alias assigned to the `develop` branch.
        """
        if self.uses_git_flow():
            return self.git_flow.develop_branch
        return 'develop'

    def get_default_branch(self) -> str:
        """Obtain the name of the branch that aliases the "default" branch.

        This is dependent on the flow. For instance, in the git flow we
        use the `develop` branch as default.

        Returns:
            The name/alias assigned to the default branch.
        """
        if self.uses_m_flow():
            return self.m_flow.master_branch
        if self.uses_git_flow():
            return self.git_flow.develop_branch
        return 'master'

    def verify_version(
        self,
        gh_latest: str,
        is_release_pr: bool,
        is_release: bool,
    ) -> Res[int]:
        """Verify that the configuration version is valid.

        Args:
            gh_latest:
                The version stored in `Github`. Checks are skipped if
                this value is empty.
            is_release_pr:
                Set to `True` if the build is done during a release pr.
            is_release:
                Set to `True` if the build is done during a release.

        Returns:
            A `OneOf` containing 0 if all is well, otherwise an `Issue`.
        """
        if not gh_latest:
            return Good(0)
        err_data = {
            'config_version': self.version,
            'gh_latest': gh_latest,
            'is_release': is_release,
            'is_release_pr': is_release_pr,
        }
        try:
            p_ver = Version(self.version)
        except Exception as ex:
            return issue('error parsing version', cause=ex, context=err_data)
        try:
            p_latest = Version(gh_latest)
        except Exception as ex:  # noqa: WPS440
            return issue('error parsing latest', cause=ex, context=err_data)
        ver_gt_latest = p_ver > p_latest
        ver_lt_latest = p_ver < p_latest
        msg: str = ''
        if is_release_pr:
            msg = _handle_release_pr(ver_gt_latest)
        elif is_release:
            msg = _handle_release(ver_gt_latest)
        else:
            msg = _handle_non_release(ver_lt_latest, ver_gt_latest)
        return issue(msg, context=err_data) if msg else Good(0)


def get_m_filename(m_dir: str) -> Res[str]:
    """Obtain the path to the m configuration file.

    Args:
        m_dir: The directory with the m configuration.

    Returns:
        The name of the configuration file or an issue if it doesn't exist.
    """
    filenames = (f'{m_dir}/m.yaml', f'{m_dir}/m.yml', f'{m_dir}/m.json')
    for filename in filenames:
        if Path(filename).exists():
            return Good(filename)
    return issue('m_file not found', context={
        'm_dir': m_dir,
        'valid_m_files': filenames,
    })


def read_config(m_dir: str) -> Res[Config]:
    """Read an m configuration file.

    Args:
        m_dir: Directory containing `m.json`.

    Returns:
        A `OneOf` containing the `m` configuration or an `Issue`.
    """
    return one_of(lambda: [
        Config(m_dir=m_dir, **m_cfg)
        for m_filename in get_m_filename(m_dir)
        for m_cfg in yaml_fp.read_yson(m_filename)
    ]).flat_map_bad(hone('read_config failure'))



import json as builtin_json
import os
from dataclasses import dataclass
from enum import Enum
from http import HTTPStatus
from http import client as httplib
from typing import Any, Mapping, Optional
from urllib.parse import urlparse

from . import Issue, issue, one_of
from .fp import Good, OneOf
from .json import parse_json


class HttpMethod(str, Enum):  # noqa: WPS600
    """Python 3.10 does not have this enum.

    https://docs.python.org/3/library/http.html#http-methods
    """

    get = 'GET'
    post = 'POST'
    put = 'PUT'

    def __str__(self) -> str:
        """Return the value of the enum.

        Returns:
            The value of the enum.
        """
        return self.value


@dataclass
class FetchedResponse:
    """Result object for fetch_response."""

    response: httplib.HTTPResponse
    body: str


def _get_connection(protocol: str, hostname: str) -> httplib.HTTPConnection:
    if protocol == 'https':
        return httplib.HTTPSConnection(hostname)  # noqa: S309
    return httplib.HTTPConnection(hostname)


def fetch_response(
    url: str,
    headers: Mapping[str, str],
    method: HttpMethod = HttpMethod.get,
    body: Optional[str] = None,
) -> OneOf[Issue, FetchedResponse]:
    """Send an http(s) request.

    Args:
        url:
            The url to request.
        headers:
            The headers for the request. By default it sets the `user-agent`
            to "m".
        method:
            The request method type. Defaults to `GET`.
        body:
            The body of the request.

    Returns:
        A `OneOf` containing the response object from the server or an Issue.
    """
    parts = urlparse(url)
    protocol, hostname, path = [parts.scheme, parts.netloc, parts.path]
    path = f'{path}?{parts.query}' if parts.query else path
    fetch_headers = {'user-agent': 'm', **headers}
    ctxt: dict[str, str] = {'url': f'{hostname}{path}', 'method': f'{method}'}
    if body:
        fetch_headers['content-length'] = str(len(body))
        if 'DEBUG_HTTP_INCLUDE_BODY' in os.environ:
            ctxt['body'] = body
    connection = _get_connection(protocol, hostname)
    # See the next link for explanation disabling WPS440:
    #  https://github.com/wemake-services/wemake-python-styleguide/issues/1416
    try:
        connection.request(f'{method}', path, body, fetch_headers)
    except Exception as ex:
        return issue(f'{protocol} request failure', cause=ex, context=ctxt)
    try:
        res = connection.getresponse()
    except Exception as ex:  # noqa: WPS440
        return issue(f'{protocol} response failure', cause=ex, context=ctxt)
    try:
        res_body = res.read().decode()
    except Exception as ex:  # noqa: WPS440
        return issue(f'{protocol} read failure', cause=ex, context=ctxt)
    code = res.getcode()
    if HTTPStatus.OK <= code < HTTPStatus.MULTIPLE_CHOICES:
        return Good(FetchedResponse(
            response=res,
            body=res_body,
        ))
    return issue(
        f'{protocol} request failure ({code})',
        context={
            'body': body,
            'code': code,
            'res_body': res_body,
            **ctxt,
        },
    )


def fetch(
    url: str,
    headers: Mapping[str, str],
    method: HttpMethod = HttpMethod.get,
    body: Optional[str] = None,
) -> OneOf[Issue, str]:
    """Send an http(s) request.

    Args:
        url:
            The url to request.
        headers:
            The headers for the request. By default it sets the `user-agent`
            to "m".
        method:
            The request method type. Defaults to `GET`.
        body:
            The body of the request.

    Returns:
        A `OneOf` containing the raw response from the server or an Issue.
    """
    return one_of(lambda: [
        fetch_res.body
        for fetch_res in fetch_response(url, headers, method, body)
    ])


def fetch_json(
    url: str,
    headers: Mapping[str, str],
    method: HttpMethod = HttpMethod.get,
    body_json: Any = None,
) -> OneOf[Issue, Any]:
    """Specialized fetch to deal with json data.

    Args:
        url:
            The url to request.
        headers:
            Additional headers for the request. By default it will add
            proper accept and content-type headers for json requests.
        method:
            The request method type. Defaults to `GET`.
        body_json:
            The data to send to the server (python object).

    Returns:
        A `OneOf` containing a json parsed response from the server or an
        Issue.
    """
    body = builtin_json.dumps(body_json) if body_json else None
    fetch_headers = {
        'accept': 'application/json',
        'content-type': 'application/json',
        **headers,
    }
    return one_of(lambda: [
        response
        for payload in fetch(url, fetch_headers, method, body)
        for response in parse_json(payload)
    ])



import sys
import time
from functools import partial
from typing import Callable

from m.core import Good, Issue, Res, issue, one_of
from m.github.api import merge_pr
from m.github.cli import get_latest_release
from m.github.graphql.enums import MergeableState
from m.github.graphql.queries.branch_prs import PullRequest
from m.github.graphql.queries.branch_prs import fetch as fetch_branch_prs
from m.log import Logger

from m import git

from .config import Config, read_config
from .release_utils import assert_branch

logger = Logger('m.ci.start_release')
_checks_per_line = 30


def inspect_prs(prs: list[PullRequest]) -> Res[None]:
    """Inspect the release pull requests.

    Attempts to warn the users of possible issues that may be encountered.

    Args:
        prs: The list of pull requests to merge.

    Returns:
        An issue if the user decides to cancel, None otherwise.
    """
    if not prs:
        return issue('no prs associated with the release/hotfix branch')
    conflicting = [
        pr
        for pr in prs
        if pr.mergeable == MergeableState.conflicting
    ]
    if conflicting:
        return issue('found conflicting prs', context={
            'prs': {pr.number: pr.url for pr in conflicting},
            'suggestion': 'check out the pull requests',
        })
    # Nice to have: handle code reviews - warn if there are requested changes
    # To introduce this we need to prompt the user if we should proceed to
    # merge the pr even when there are pending reviews. This is mainly
    # done for admins that may be able to merge without approvals.
    return Good(None)


def wait_until(
    predicate: Callable[[], Res[bool]],
) -> Res[None]:
    """Sleep until the predicate function returns False.

    Will print a `.` (dot) every 10 seconds thus every 6 dots
    denote a minute.

    Args:
        predicate: function returning an Issue or a boolean

    Returns:
        An issue or None.
    """
    should_loop = predicate()
    counter = 0
    indent = '       '
    sys.stdout.write(indent)
    while not should_loop.is_bad and should_loop.value is True:
        time.sleep(10)
        counter += 1
        modifier = ''
        if counter % 6 == 0:
            modifier = '    '
        if counter % _checks_per_line == 0:
            modifier = f'\n{indent}'
        sys.stdout.write(f'.{modifier}')
        sys.stdout.flush()
        should_loop = predicate()
    sys.stdout.write('\n')
    return should_loop.map(lambda _: None)


def merge_prs(
    gh_token: str,
    config: Config,
    prs: list[PullRequest],
    target_ver: str,
) -> Res[None]:
    """Merge the given prs.

    Args:
        gh_token: The GITHUB_TOKEN to use to make api calls to Github.
        config: The m configuration.
        prs: The pull requests to merge.
        target_ver: The version to release.

    Returns:
        An issue if there is a problem while merging or None if successful.
    """
    master_branch = config.get_master_branch()
    first_index = 0 if prs[0].base_ref_name == master_branch else 1
    second_index = 0 if first_index == 1 else 1
    first_pr = prs[first_index]
    second_pr = None
    if len(prs) == 2:
        second_pr = prs[second_index]

    first_merged_result: Res[None]
    if first_pr.merged or first_pr.closed:
        msg = f'{master_branch} branch pr already merged/closed'
        logger.warning(msg, context={
            'pr_info': first_pr.model_dump(),
        })
        first_merged_result = Good(None)
    else:
        msg = f'merged pr{first_pr.number} to {master_branch}'
        first_merged_result = merge_pr(
            gh_token,
            config.owner,
            config.repo,
            first_pr.number,
            None,
        ).map(lambda res: logger.info(msg, res)).map(lambda _: None)

    return one_of(lambda: [
        None
        for _ in first_merged_result
        for _ in _merge_second_pr(gh_token, config, second_pr, target_ver)
    ])


def _not_released(
    token: str,
    owner: str,
    repo: str,
    ver: str,
) -> Res[bool]:
    latest = get_latest_release(token, owner, repo)
    return latest.map(lambda current_ver: current_ver != ver)


def _merge_second_pr(
    gh_token: str,
    config: Config,
    pr: PullRequest | None,
    target_ver: str,
) -> Res[None]:
    if not pr:
        return Good(None)
    base_ref = pr.base_ref_name
    if pr.merged or pr.closed:
        logger.warning(f'{base_ref} branch pr already merged/closed', context={
            'pr_info': pr.model_dump(),
        })
        return Good(None)
    ver = target_ver
    owner = config.owner
    repo = config.repo
    pr_num = pr.number
    not_released = partial(_not_released, gh_token, owner, repo, ver)
    logger.info(f'checking every 10 seconds until {ver} is released')
    return one_of(lambda: [
        None
        for _ in wait_until(not_released)
        for res in merge_pr(gh_token, owner, repo, pr_num, None)
        for _ in logger.info(f'merged pr{pr_num} to {base_ref}', res)
    ])


def _switch_branch(branch: str, checkout: str) -> Res[None]:
    return logger.info(f'release complete - switching to "{branch}" branch', {
        'git': f'{checkout}\n',
    }).map(lambda _: None)


def _pull_branch(pull: str) -> Res[None]:
    return logger.info('pulling the default branch', {
        'git': f'{pull}\n',
    }).map(lambda _: None)


def end_release(gh_token: str) -> Res[None]:
    """End the release process.

    Args:
        gh_token: The GITHUB_TOKEN to use to make api calls to Github.

    Returns:
        None if successful, otherwise an issue.
    """
    return one_of(lambda: [
        None
        for branch in git.get_branch()
        for _, target_ver in assert_branch(branch, 'end')
        for config in read_config('m')
        for prs in fetch_branch_prs(
            gh_token,
            config.owner,
            config.repo,
            branch,
        )
        for _ in inspect_prs(prs)
        for _ in merge_prs(gh_token, config, prs, target_ver)
        for default_branch in Good[Issue, str](config.get_default_branch())
        for checkout in git.checkout_branch(default_branch, create=False)
        for _ in _switch_branch(default_branch, checkout)
        for pull_output in git.pull()
        for _ in _pull_branch(pull_output)
    ])



import re
from typing import Any, Optional

from m.core import Good, Issue, OneOf, non_issue, one_of

from ..core.json import get
from .ci_dataclasses import (
    AssociatedPullRequest,
    Author,
    Commit,
    CommitInfo,
    GithubCiRunInfo,
    PullRequest,
    Release,
)
from .ci_graph_queries import LATEST_RELEASE, PULL_REQUEST, commit_query
from .graphql import api


def create_ci_query(
    pr_number: int | None = None,
    include_commit: bool = False,
    include_release: bool = False,
) -> str:
    """Create github graphql query.

    The information provided is done via presets.

    Args:
        pr_number: If included, it will add pull request information.
        include_commit: If true, include commit information.
        include_release: If true, include release information.

    Returns:
        A string with the graphql query.
    """
    query_items: list[str] = []
    query_params: list[str] = ['$owner: String!', '$repo: String!']
    if include_commit:
        include_pr = pr_number is None
        query_items.append(commit_query(include_pr, include_author=True))
        query_params.append('$sha: String')
    if pr_number:
        query_items.append(PULL_REQUEST)
        query_params.append('$pr: Int!')
        query_params.append('$fc: Int!')
    if include_release:
        query_items.append(LATEST_RELEASE)
    items_str = '\n'.join(query_items)
    params_str = ', '.join(query_params)
    return f"""query ({params_str}) {{
      repository(owner:$owner, name:$repo) {{
        {items_str}
      }}
    }}"""


def _parse_commit_message(msg: str, sha: str) -> str:
    match = re.match('^Merge (.*) into (.*)$', msg)
    if match:
        return match.groups()[0]
    return sha


def get_build_sha(
    token: str,
    owner: str,
    repo: str,
    sha: str,
) -> OneOf[Issue, str]:
    """Obtain the commit sha in a pull request.

    When building prs, we are not given the actual sha of the commit.

    Instead, we get the sha of the merge commit. This will give us the
    sha that we are looking for.

    Args:
        token: A Github PAT.
        owner: The owner of the repo.
        repo: The name of the repo.
        sha: The merge sha provided by github.

    Returns:
        The actual sha of the commit.
    """
    query_params = ['$owner: String!', '$repo: String!', '$sha: String!']
    params_str = ', '.join(query_params)
    commit_query_str = commit_query(include_pr=False, include_author=False)
    query = f"""query ({params_str}) {{
      repository(owner:$owner, name:$repo) {{
        {commit_query_str}
      }}
    }}"""
    variables = {
        'owner': owner,
        'repo': repo,
        'sha': sha,
    }
    return one_of(
        lambda: [
            _parse_commit_message(commit_msg, sha)
            for res in api.graphql(token, query, variables)
            for commit_msg in get(res, 'repository.commit.message')
        ],
    )


def get_raw_ci_run_info(
    token: str,
    commit_info: CommitInfo,
    pr_number: int | None,
    file_count: int,
    include_release: bool,
    get_sha: bool = True,
) -> OneOf[Issue, Any]:
    """Retrieve the information of the given Github PR.

    Args:
        token: A Github PAT.
        commit_info: The owner, repo and sha.
        pr_number: The pull request number.
        file_count:  The maximum number of files in the pr to retrieve.
        include_release: If true it will provide release information.
        get_sha: If true, it will obtain the actual sha of the commit.

    Returns:
        The Github payload with the raw information.
    """
    query = create_ci_query(
        pr_number,
        include_commit=True,
        include_release=include_release,
    )
    owner, repo, sha = [commit_info.owner, commit_info.repo, commit_info.sha]
    variables = {'owner': owner, 'repo': repo, 'sha': sha, 'fc': file_count}
    if pr_number:
        variables['pr'] = pr_number
    if get_sha:
        build_sha = get_build_sha(token, owner, repo, sha)
        if build_sha.is_bad:
            return build_sha
        variables['sha'] = non_issue(build_sha)
    return one_of(lambda: [
        repo_data
        for res in api.graphql(token, query, variables)
        for repo_data in get(res, 'repository')
    ])


def _get_release(raw: Any) -> OneOf[Issue, Optional[Release]]:
    releases = raw.get('releases')
    if releases:
        nodes = releases.get('nodes')
        if nodes:
            node = nodes[0]
            return Good(
                Release(
                    name=node['name'],
                    tag_name=node['tagName'],
                    published_at=node['publishedAt'],
                ),
            )
    return Good(None)


def _get_commit(owner: str, repo: str, raw: Any) -> OneOf[Issue, Commit]:
    commit = raw['commit']
    sha = commit['oid']
    commit_info = Commit(
        author_login=get(commit, 'author.user.login').get_or_else(''),
        short_sha=sha[:7],
        sha=sha,
        message=commit['message'],
        url=f'https://github.com/{owner}/{repo}/commit/{sha}',
    )
    nodes = get(commit, 'associatedPullRequests.nodes').get_or_else([])
    if nodes:
        pr = nodes[0]
        author = pr.get('author')
        commit_info.associated_pull_request = AssociatedPullRequest(
            author=Author(**author),
            merged=pr.get('merged'),
            pr_number=pr.get('number'),
            target_branch=pr.get('baseRefName'),
            target_sha=pr.get('baseRefOid'),
            pr_branch=pr.get('headRefName'),
            pr_sha=pr.get('headRefOid'),
            title=pr.get('title'),
            body=pr.get('body'),
        )
    return Good(commit_info)


def _get_pull_request(
    raw: Any,
    pr_number: Optional[int],
) -> OneOf[Issue, Optional[PullRequest]]:
    pr = raw.get('pullRequest')
    if not pr:
        return Good(None)
    author = pr.get('author', {})
    return Good(
        PullRequest(
            author=Author(**author),
            pr_number=pr_number or 0,
            pr_branch=pr.get('headRefName'),
            target_branch=pr.get('baseRefName'),
            target_sha=pr.get('baseRefOid'),
            url=pr.get('url'),
            title=pr.get('title'),
            body=pr.get('body'),
            file_count=pr.get('files', {}).get('totalCount', 0),
            files=[x['path'] for x in pr.get('files', {}).get('nodes', [])],
            is_draft=pr.get('isDraft'),
        ),
    )


def get_ci_run_info(
    token: str,
    commit_info: CommitInfo,
    pr_number: int | None,
    file_count: int,
    include_release: bool,
) -> OneOf[Issue, GithubCiRunInfo]:
    """Transform the result from get_raw_ci_run_info to a GithubCiRunInfo.

    Args:
        token: A Github PAT.
        commit_info: An instance of a commit info.
        pr_number: An optional pull request number.
        file_count: The maximum number of files to report.
        include_release: If true it will provide release information.

    Returns:
        If successful, a `GithubCiRunInfo` instance.
    """
    raw_res = get_raw_ci_run_info(
        token,
        commit_info,
        pr_number,
        file_count,
        include_release,
    )
    return one_of(
        lambda: [
            GithubCiRunInfo(commit=commit, pull_request=pr, release=release)
            for raw in raw_res
            for release in _get_release(raw)
            for commit in _get_commit(commit_info.owner, commit_info.repo, raw)
            for pr in _get_pull_request(raw, pr_number)
        ],
    )


def compare_sha_url(owner: str, repo: str, prev: str, nxt: str) -> str:
    """Generate a url to compare two sha/tags in a github repo.

    Args:
        owner: The repo owner.
        repo: The repo name.
        prev: The previous sha.
        nxt: The next sha.

    Returns:
        A url comparing two shas.
    """
    return f'https://github.com/{owner}/{repo}/compare/{prev}...{nxt}'



import argparse
from functools import partial
from inspect import cleandoc
from typing import Callable
from warnings import warn

from pydantic import BaseModel
from pydantic.fields import FieldInfo

from .misc import params_count
from .parsers import boolean, positional, proxy, remainder, standard
from .types import CommandInputs, FuncArgs


def _parse_field(name: str, field: FieldInfo) -> FuncArgs:
    extras = field.json_schema_extra
    if not isinstance(extras, dict):  # pragma: no cover
        # We can go out of our way to not make it a dict but then that's on the
        # developer. Won't handle the case that json_schema_extra is a function.
        not_supported = 'json_schema_extra only supported as a dict'
        raise NotImplementedError(not_supported)
    if positional.should_handle(extras):
        return positional.handle_field(name, field)
    if field.annotation is bool:
        return boolean.handle_field(name, field)
    if proxy.should_handle(extras):
        return proxy.handle_field(extras)
    if remainder.should_handle(extras):
        return remainder.handle_field(name, field)
    return standard.handle_field(name, field)


def add_model(
    parser: argparse.ArgumentParser,
    model: type[BaseModel],
) -> None:
    """Add a pydantic model to an argparse ArgumentParser.

    Args:
        parser: The ArgumentParser instance.
        model: The pydantic model declaring the cli options.
    """
    parser.description = cleandoc(model.__doc__ or '')
    parser.formatter_class = argparse.RawTextHelpFormatter
    for name, field in model.model_fields.items():
        arg_inputs = _parse_field(name, field)
        parser.add_argument(*arg_inputs.args, **arg_inputs.kwargs)


def _run_wrapper(
    run_func: Callable[..., int],
    cmd_inputs: CommandInputs,
    cmd_name: str,
    arg: argparse.Namespace | None,
    parser: argparse._SubParsersAction | None,  # noqa: WPS437
) -> int:
    if parser:
        sub_parser = parser.add_parser(cmd_name, help=cmd_inputs.help)
        add_model(sub_parser, cmd_inputs.model)
        return 0
    if isinstance(arg, argparse.Namespace):
        arg_dict = arg.__dict__
        opt = cmd_inputs.model.model_validate(arg_dict)
        len_run_params = params_count(run_func)
        args = [opt, arg][:len_run_params]
        return run_func(*args)
    # Should not be reachable during a normal run - only checking
    # to make sure we provide at least one argument.
    raise NotImplementedError(  # pragma: no cover
        'm dev error: provide either arg or parser',
    )


def _handle_decorated_func(cmd_inputs: CommandInputs, func: Callable):
    return partial(_run_wrapper, func, cmd_inputs)


def command(
    *,
    help: str,  # noqa: WPS125
    model: type[BaseModel],
    name: str | None = None,
) -> partial[partial[int]]:
    """Apply a decorator to the `run` function to make it into a command.

    Args:
        name: The command name.
        help: A short description of the command.
        model: A pydantic model to describe the cli arguments.

    Returns:
        A transformed run function aware of the arguments model.
    """
    # m no longer uses the name argument but we keep it for now
    if name:  # pragma: no cover
        warn('`name` is no longer needed, please remove it', DeprecationWarning)
    return partial(_handle_decorated_func, CommandInputs(help, model))



from m.core import yaml
from m.pydantic import KebabModel
from pydantic import BaseModel
from pydantic.fields import PydanticUndefined

from .actions import Action
from .misc import MetadataOutput

TEMPLATE = """\
# AUTOGENERATED FILE - Do not update by hand!
# Edit {py_path}/{module_name}.py
#   then run `m github actions [python_file]` to update

name: {name}
description: {description}

inputs:{inputs}
outputs:{outputs}
runs:
  using: composite
  steps:
    {steps}
"""
EMPTY_DICT = '{}'  # noqa: P103 - not a format string
InputInfo = dict[str, str | bool]  # container for pydantic model properties


class TemplateVars(BaseModel):
    """Template variables."""

    py_path: str

    module_name: str

    name: str

    description: str

    inputs: str

    outputs: str

    steps: str


def _indent(text: str, num: int) -> str:
    spaces = '  ' * num
    return text.replace('\n', f'\n{spaces}', -1)


class ActionFile(BaseModel):
    """Helper class to write the action file."""

    py_path: str

    module_name: str

    name: str

    description: str

    action: Action

    inputs: type[KebabModel]

    outputs: dict[str, MetadataOutput]

    python_path: str

    available_outputs: dict[str, str]

    def build_inputs(self: 'ActionFile') -> str:
        """Generate the inputs for the action.

        Returns:
            A string to add to the Github action.
        """
        all_inputs: dict[str, InputInfo] = {}
        for prop_name, prop_info in self.inputs.model_fields.items():
            input_info: InputInfo = {
                'description': prop_info.description or '',
                'required': prop_info.is_required(),
            }
            default_val = prop_info.get_default()
            if default_val is not PydanticUndefined:
                input_info['default'] = default_val
            all_inputs[prop_info.alias or prop_name] = input_info
        if not all_inputs:
            return f' {EMPTY_DICT}'
        str_inputs = _indent(yaml.dumps(all_inputs), 1)
        return f'\n  {str_inputs}'

    def build_outputs(
        self: 'ActionFile',
        outputs: dict[str, MetadataOutput],
    ) -> str:
        """Generate the outputs for the action.

        Args:
            outputs: The outputs to add to the action.

        Returns:
            A string to add to the Github action.
        """
        if not outputs:
            return f' {EMPTY_DICT}'
        raw_data = {
            output_name.replace('_', '-', -1): output_val.model_dump()
            for output_name, output_val in outputs.items()
        }
        str_outputs = _indent(yaml.dumps(raw_data), 1)
        return f'\n  {str_outputs}'

    def build_steps_str(self: 'ActionFile') -> str:
        """Generate a github action str for the build steps.

        Returns:
            A string to add to the Github action.
        """
        lines: list[str] = [
            step.to_str(self.python_path, self.available_outputs)
            for step in self.action.steps
        ]
        return _indent('\n'.join(lines), 2)

    def __str__(self: 'ActionFile') -> str:
        """Stringify the action file.

        Returns:
            The github action.
        """
        template_vars = TemplateVars(
            py_path=self.py_path,
            module_name=self.module_name.replace('.', '/', -1),
            name=self.name,
            description=self.description,
            inputs=self.build_inputs(),
            outputs=self.build_outputs(self.outputs),
            steps=self.build_steps_str(),
        )
        file_content = TEMPLATE.format(**template_vars.model_dump())
        return '\n'.join([
            line.rstrip()
            for line in file_content.split('\n')
        ])



from pydantic import BaseModel

from .args import Arg, ArgProxy, Meta, RemainderArgs
from .cli import cli_integration_token, exec_cli, run_cli, run_main
from .engine.argparse import command
from .engine.sys import (
    cli_commands,
    command_group,
    import_cli_commands,
    merge_cli_commands,
    subcommands,
)
from .engine.types import (
    CliCommands,
    CommandModule,
    FuncArgs,
    MetaModule,
    add_arg,
)
from .handlers import (
    create_issue_handler,
    create_json_handler,
    create_yaml_handler,
)
from .validators import (
    env_var,
    env_var_or_empty,
    validate_file_exists,
    validate_json_payload,
    validate_payload,
)

# using as barrel file to export convenience functions
# pydantic.BaseModel has been re-exported as a convenience to not
# have so many imports in command files.
__all__ = (  # noqa: WPS410
    'cli_integration_token',
    'cli_commands',
    'subcommands',
    'command_group',
    'create_issue_handler',
    'create_json_handler',
    'create_yaml_handler',
    'command',
    'import_cli_commands',
    'add_arg',
    'validate_json_payload',
    'validate_payload',
    'validate_file_exists',
    'run_main',
    'exec_cli',
    'merge_cli_commands',
    'run_cli',
    'FuncArgs',
    'Arg',
    'ArgProxy',
    'RemainderArgs',
    'BaseModel',
    'Meta',
    'MetaModule',
    'CommandModule',
    'CliCommands',
    'env_var',
    'env_var_or_empty',
)



import re
from functools import partial
from textwrap import dedent
from typing import Callable, Generic, TypeVar, cast

from m.core import Good, Res, issue
from m.log import Logger
from m.pydantic import KebabModel
from pydantic import BaseModel
from pydantic.fields import Field

from .misc import (
    ActionOutputs,
    InputOutputs,
    MetadataOutput,
    OutputField,
    get_inputs_outputs,
    input_env,
    map_args,
    verify_inputs,
)

InputModel = TypeVar('InputModel', bound=KebabModel)
OutputModel = TypeVar('OutputModel', bound=KebabModel)
logger = Logger('m.github.actions')


def _replace_values(available_values: dict[str, str], match: re.Match) -> str:
    step_id, arg_name = match.groups()
    key = f'{step_id}.{arg_name}'
    if key not in available_values:
        logger.warning(f'Unknown step value: {key}')
    return available_values.get(key, key)


def _run_if(condition: str | None, available_values: dict[str, str]) -> str:
    run_if = ''
    if condition:
        replace = partial(_replace_values, available_values)
        with_replacements = re.sub(
            r'\$([\w-]+).([\w-]+)',
            replace,
            condition,
            flags=re.I | re.M,
        )
        run_if = f'\n  if: {with_replacements}'
    return run_if


class RunStep(BaseModel, Generic[InputModel, OutputModel]):
    """Model used to define a "run" step in a Github action.

    The `id` is important because this is how we will be able to refer to the
    outputs generated by the step. Say our action called other external actions
    such as the cache action. Then if we wanted to pass one of the outputs to
    the cache action we would have to get a handle on the step.

    ```python
    args=SomeInput(some_arg='my_run_step_id.some_output')
    ```

    The `args` should leverage the help of their own input models. All they
    require is that we provide the handle to other outputs from other steps or
    some other values we wish to pass.

    Experiment with different values to see what `action.yaml` generates.
    """

    id: str = Field(description='The id of the step.')

    run_if: str | None = Field(
        default=None,
        description='The condition to run the step.',
    )

    run: Callable[[InputModel], Res[OutputModel]] = Field(
        description='The function passed to [`run_action`][m.github.actions.api.run_action].',
    )

    args: InputModel | None = Field(
        description='The arguments to pass to the run function.',
    )

    def get_inputs_outputs(self: 'RunStep') -> InputOutputs:
        """Get the inputs and outputs for the step.

        Returns:
            A tuple of the inputs and outputs.
        """
        return get_inputs_outputs(self.run)

    def to_str(
        self: 'RunStep',
        python_path: str,
        available_values: dict[str, str],
    ) -> str:
        """Generate a string to use in the Github Action.

        Args:
            python_path: The path to the python module.
            available_values: The values that are available to the step.

        Returns:
            A string to add to the Github action.
        """
        template = """\
            - id: {id}{run_if}
              shell: bash{env}
              run: PYTHONPATH="$GITHUB_ACTION_PATH{py_path}" python -m {mod}
        """
        run_if = _run_if(self.run_if, available_values)
        # mypy has trouble seeing that its bound to KebabModel
        self_args = cast(KebabModel, self.args)
        all_args = self_args.model_dump() if self_args else {}
        mapped_args = map_args(all_args, available_values, input_env)
        env = ''
        if mapped_args:
            arg_lines = '\n'.join([
                f'    {key}: {env_val}'
                for key, env_val in mapped_args.items()
            ])
            env = f'\n  env:\n{arg_lines}'
        py_path = f'/{python_path}' if python_path else ''
        return dedent(template).format(
            id=self.id,
            env=env,
            run_if=run_if,
            py_path=py_path,
            mod=self.run.__module__,
        ).rstrip()


class UsesStep(BaseModel, Generic[InputModel, OutputModel]):
    """A "uses" step in a Github action.

    Model similar to [`RunStep`][m.github.actions.RunStep] but since we do
    not have access to the code that gets executed all we can do is provide the
    `uses` field and our models to describe what the action expects.

    For instance, say we wanted to use the `actions/cache@v4` action. To avoid
    having issues in the future we should create the input and output models
    manually by looking at the documentation
    <https://github.com/actions/cache#inputs>.

    ```python
    class CacheInputs(KebabModel):
        key: str = InArg(help='An explicit key for a cache entry')
        path: str = InArg(help=\"\"\"
            A list of files, directories, and wildcard patterns to cache and
            restore.
        \"\"\")
    ```

    Similarly for the outputs we can define a model. This in the long run should
    help us maintain the composite action better. It is recommended to check the
    inputs and outputs as we update action versions to ensure compatibility.
    """  # noqa: D301, D300 - Angry with slash but we need to escape them

    id: str = Field(description='The step id.')

    run_if: str | None = Field(
        default=None,
        description='The condition to run the step.',
    )

    uses: str = Field(description='A string referencing an action.')

    inputs: type[InputModel] = Field(
        description='A reference to a [KebabModel][m.pydantic.KebabModel] type.',
    )

    outputs: type[OutputModel] = Field(
        description='A reference to a [KebabModel][m.pydantic.KebabModel] type.',
    )

    args: InputModel | None = Field(
        description='The arguments to pass to the action',
    )

    def get_inputs_outputs(self: 'UsesStep') -> InputOutputs:
        """Get the inputs and outputs for the step.

        Returns:
            A tuple of the inputs and outputs.
        """
        return self.inputs, self.outputs

    def to_str(
        self: 'UsesStep',
        _python_path: str,
        available_values: dict[str, str],
    ) -> str:
        """Generate a string to use in the Github Action.

        Args:
            _python_path: The path to the python module.
            available_values: The values that are available to the step.

        Returns:
            A string to add to the Github action.
        """
        template = """\
            - id: {id}{run_if}
              uses: {uses}{env}
        """
        run_if = _run_if(self.run_if, available_values)
        # mypy has trouble seeing that its bound to KebabModel
        self_args = cast(KebabModel, self.args)
        all_args = self_args.model_dump() if self_args else {}
        mapped_args = map_args(
            all_args,
            available_values,
            lambda x: x.replace('_', '-', -1),
        )
        env = ''
        if mapped_args:
            arg_lines = '\n'.join([
                f'    {key}: {env_val}'
                for key, env_val in mapped_args.items()
            ])
            env = f'\n  with:\n{arg_lines}'
        return dedent(template).format(
            id=self.id,
            run_if=run_if,
            uses=self.uses,
            env=env,
        ).rstrip()


class Action(BaseModel):
    """A Github action.

    The main object to help us define the `actions.yaml` file. Each repository can
    declare several actions. For this reason we can declare the `actions` object as
    a single `Action` or a list of them.

    We could technically do everything in one single step but when we use external
    actions then we are forced to split the step. Note that we have no way of
    having the `if` fields in each step. This is because we are encouraged to handle
    that logic in each of the steps we declare. We have full control of the output to
    `stdout` and `stderr` here.
    """

    file_path: str = Field(description='The full path to the `action.yaml` file.')

    name: str = Field(description='The name of the action.')

    description: str = Field(description='Short description for the action.')

    inputs: type[KebabModel] | None = Field(description="""
        A model describing the inputs for the action. If the action does not
        need inputs then provide `None`. It is important to be explicit in this step
        so that we do not wonder why we cannot use `inputs.[argname]` when we declare a
        step.
    """)

    steps: list[RunStep | UsesStep] = Field(
        description='The steps for the action',
    )

    def gather_outputs(self: 'Action') -> Res[ActionOutputs]:
        """Obtain a tuple with the action outputs and all steps outputs.

        The steps outputs is a dictionary that maps keys of the form
        [step_id].[step_output_arg] to the output of another step, action
        input or some other value that should be used as input.

        This function validates that all the keys are valid.

        Returns:
            A tuple with the outputs if successful, otherwise an issue.
        """
        action_inputs = self.inputs or KebabModel
        available_outputs: dict[str, str] = {
            f'inputs.{name}': f'inputs.{arg_info.alias}'
            for name, arg_info in action_inputs.model_fields.items()
        }

        outputs: dict[str, MetadataOutput] = {}
        all_issues: dict[str, list[dict]] = {}
        for step in self.steps:
            input_model, output_model = step.get_inputs_outputs()
            output_fields = OutputField.create(step.id, output_model)
            # mypy has trouble seeing that its bound to KebabModel
            self_args = cast(KebabModel, step.args)
            all_args = self_args.model_dump() if self_args else {}
            issues = verify_inputs(input_model, all_args, available_outputs)
            if issues:
                all_issues[step.id] = [
                    iss.model_dump(exclude_none=True)
                    for iss in issues
                ]
            for name, field in output_fields.items():
                short_name = field.short_ref_name
                available_outputs[short_name] = field.full_ref_name
                if field.is_export:
                    outputs[name] = field.get_metadata_output()
        if all_issues:
            return issue('step_inputs_failure', context=all_issues)
        return Good((outputs, available_outputs))



from pathlib import Path

from pydantic import BaseModel


class FileNames(BaseModel):
    """Name of files that need to be written."""

    m_dir: str
    local_dir: str
    ci_dir: str
    manifests_dir: str
    makefile: str
    gh_workflow: str

    @classmethod
    def create_instance(cls: type['FileNames'], m_dir: str) -> 'FileNames':
        """Create and instance of FileNames.

        Args:
            m_dir: The m directory.

        Returns:
            An instance of FileNames.
        """
        gh_dir = Path(f'{m_dir}/../.github/workflows')
        gh_dir.mkdir(parents=True, exist_ok=True)
        return FileNames(
            m_dir=m_dir,
            local_dir=f'{m_dir}/.m/blueprints/local',
            ci_dir=f'{m_dir}/.m/blueprints/ci',
            manifests_dir=f'{m_dir}/.m/blueprints/ci/manifests',
            makefile=f'{m_dir}/../Makefile',
            gh_workflow=f'{m_dir}/../.github/workflows/m.yaml',
        )

    def local_file(self: 'FileNames', filename: str) -> str:
        """Generate the name of a local script.

        Args:
            filename: The name of the file.

        Returns:
            The full name of the file.
        """
        return f'{self.local_dir}/{filename}'



from pydantic import BaseModel

from .shell_command import ShellCommand


class DockerBuild(BaseModel):
    """Representation of a `docker build` command.

    See https://docs.docker.com/engine/reference/commandline/build
    """

    # Add a custom host-to-IP mapping (host:ip)
    add_host: list[str] | None = None

    # Set build-time variables
    build_arg: list[str] | None = None

    # Images to consider as cache sources
    cache_from: str | None = None

    # Optional parent cgroup for the container
    cgroup_parent: str | None = None

    # Compress the build context using gzip
    compress: bool | None = None

    # Limit the CPU CFS (Completely Fair Scheduler) period
    cpu_period: int | None = None

    # Limit the CPU CFS (Completely Fair Scheduler) quota
    cpu_quota: int | None = None

    # CPU shares (relative weight)
    cpu_shares: int | None = None

    # CPUs in which to allow execution (0-3, 0,1)
    cpuset_cpus: str | None = None

    # MEMs in which to allow execution (0-3, 0,1)
    cpuset_mems: str | None = None

    # Skip image verification (default true)
    disable_content_trust: bool | None = None

    # Name of the Dockerfile (Default is 'PATH/Dockerfile')
    file: str | None = None  # noqa: WPS110 - docker build optional argument

    # Always remove intermediate containers
    force_rm: bool | None = None

    # Write the image ID to the file
    iidfile: str | None = None

    # Container isolation technology
    isolation: str | None = None

    # Set metadata for an image
    label: list[str] | None = None

    # Memory limit
    memory: str | None = None

    # Swap limit equal to memory plus swap: '-1' to enable unlimited swap
    memory_swap: str | None = None

    # Set the networking mode for the RUN instructions during build (default "default")
    network: str | None = None

    # Do not use cache when building the image
    no_cache: bool | None = None

    # Always attempt to pull a newer version of the image
    pull: bool | None = None

    # Suppress the build output and print image ID on success
    quiet: bool | None = None

    # Remove intermediate containers after a successful build (default true)
    rm: bool | None = None

    # Security options
    security_opt: str | None = None

    # Available with DOCKER_BUILDKIT=1
    #   see https://pythonspeed.com/articles/docker-build-secrets
    # Use as `--secret id=ENVVAR` in the docker build command.
    # Then in the `RUN` statement in the docker file do `RUN --mount=type=secret,id=ENVVAR`.
    # Then you may access that value by doing `ENVVAR=$(cat /run/secrets/ENVVAR)`.
    # Best to use that in a script that that executes inside the docker file.
    secret: list[str] | None = None

    # Available with DOCKER_BUILDKIT=1
    progress: str | None = None

    # Size of /dev/shm
    shm_size: str | None = None

    # Name and optionally a tag in the 'name:tag' format
    tag: list[str] | None = None

    # Set the target build stage to build
    target: str | None = None

    # Ulimit options (default [])
    ulimit: str | None = None

    def __str__(self) -> str:
        """Convert the docker build command to a string.

        Returns:
            The docker build command.
        """
        cmd = ShellCommand(
            prog='docker build',
            positional=['.'],
            options=self.model_dump(exclude_none=True),
        )
        return str(cmd)



from enum import Enum

from pydantic import BaseModel

# Note: cannot use StrEnum because of py310


class Branches(str, Enum):  # noqa: WPS600
    """Default branches to use for the supported workflows."""

    master = 'master'
    develop = 'develop'
    release = 'release'
    hotfix = 'hotfix'

    def __str__(self: 'Branches') -> str:
        """Return the string representation of the Branch.

        Returns:
            The string representation of the Branch.
        """
        return self.value


class Workflow(str, Enum):  # noqa: WPS600
    """Supported workflows."""

    git_flow = 'git_flow'
    m_flow = 'm_flow'
    free_flow = 'free_flow'

    def __str__(self: 'Workflow') -> str:
        """Return the string representation of the workflow.

        Returns:
            The string representation of the workflow.
        """
        return self.value


class GitFlowConfig(BaseModel):
    """An object mapping branches for the git_flow workflow."""

    master_branch: str | Branches = Branches.master
    develop_branch: str | Branches = Branches.develop
    release_prefix: str | Branches = Branches.release
    hotfix_prefix: str | Branches = Branches.hotfix


class MFlowConfig(BaseModel):
    """An object mapping branches for the m_flow workflow."""

    master_branch: str | Branches = Branches.master
    release_prefix: str | Branches = Branches.release
    hotfix_prefix: str | Branches = Branches.hotfix



from m.cli import BaseModel, command


class Arguments(BaseModel):
    """Command line prompter.

    The goal is to display useful git information in the shell prompt.
    """


@command(
    help='print out a shell prompt with git info',
    model=Arguments,
)
def run() -> int:
    import sys

    from m.devcontainer.prompter import prompter

    sys.stdout.write(prompter())
    return 0



from m.cli import ArgProxy, BaseModel, command, env_var, run_main


class Arguments(BaseModel):
    """Create Github pull request(s) to review the release."""

    github_token: str = ArgProxy(
        '--github-token',
        type=env_var,
        default='GITHUB_TOKEN',
        help='Github personal access token with repo scope',
    )


@command(
    help='create release pull requests',
    model=Arguments,
)
def run(arg: Arguments):
    from m.ci.review_release import review_release

    return run_main(lambda: review_release(arg.github_token))



from m.cli import Arg, BaseModel, command


class Arguments(BaseModel):
    """Add a log message providing basic information about the devcontainer."""

    img_name: str = Arg(help='name of the docker image', required=True)
    img_version: str = Arg(help='version of the docker image', required=True)
    changelog_url: str | None = Arg(default=None, help='url to the changelog')


@command(
    help='display devcontainer information on start',
    model=Arguments,
)
def run(arg: Arguments) -> int:
    from m.devcontainer.greet import greet

    greet(arg.img_name, arg.img_version, arg.changelog_url)
    return 0



from __future__ import annotations

from typing import (
    Any,
    Callable,
    Generic,
    Iterator,
    TypeGuard,
    TypeVar,
    Union,
    cast,
)

import typing_extensions

# Note: disabling WPS110 in .flake8 for this file.
#   Reason for this, the variable name "value".
A = TypeVar('A')  # pylint: disable=invalid-name
B = TypeVar('B')  # pylint: disable=invalid-name
G = TypeVar('G')  # pylint: disable=invalid-name
K = TypeVar('K')  # pylint: disable=invalid-name
LazyArg = Union[A, Callable[[], A]]


def lazy_arg(z_arg: LazyArg[A]) -> A:
    """Return the output of calling `z_arg` if it is function.

    Otherwise param is returned.

    Args:
        z_arg: A function or a value

    Returns:
        The value.
    """
    return z_arg() if callable(z_arg) else z_arg


class StopBadIteration(Exception):  # noqa: N818 - This is for internal use
    """Store a `Bad` instance."""

    def __init__(self, bad: Any):
        """Initialize the Exception.

        Args:
            bad: The `Bad` value stored in a `OneOf`.
        """
        super().__init__()
        self.bad = bad


class Bad(Generic[B, G]):
    """The bad side of the disjoint union."""

    value: B
    is_bad = True

    def __init__(self, value: B):
        """Initialize a `Bad` instance.

        Args:
            value: The "bad" value to store in the instance.
        """
        self.value = value

    def __iter__(self) -> Iterator[G]:
        """Iterate over values of an instance.

        The intention is to raise a `StopBadIteration` exception to
        be able to break out of for loop comprehensions.

        Raises:
            StopBadIteration: If the instance has a "Bad" value.
        """
        raise StopBadIteration(self)

    def iter(self) -> Iterator[G]:
        """Shortcut to transform to a list.

        Can be used as `list(x.iter())`. It will either contain a value or be
        an empty list.

        Yields:
            The value if the instance is a "Good" value.
        """
        empty = ()
        yield from empty

    def get_or_else(self, default: LazyArg[G]) -> G:
        """Return the value if its Good or the given argument if its a Bad.

        Args:
            default: The default value in case the instance is "Bad".

        Returns:
            Either the value or the default specified by "default".
        """
        return lazy_arg(default)

    def map(self, _fct: Callable[[G], K]) -> 'Bad[B, K]':
        """Apply the function to its value if this is a `Good` instance.

        Args:
            _fct: The function to apply to the "Good" value.

        Returns:
            Itself if its a `Bad` otherwise another instance of `Good`.
        """
        return cast(Bad[B, K], self)

    def flat_map_bad(self, fct: Callable[[B], OneOf[B, G]]) -> OneOf[B, G]:
        """Apply the input function if this is a `Bad` value.

        Args:
            fct: The function to apply to the "Bad" value.

        Returns:
            Itself if its a `Good` otherwise another instance of `Bad`.
        """
        return fct(self.value)


class Good(Generic[B, G]):
    """The good side of the disjoint union."""

    value: G
    is_bad = False

    def __init__(self, value: G):
        """Initialize a `Good` instance.

        Args:
            value: The "good" value to store in the instance.
        """
        self.value = value

    def __iter__(self) -> Iterator[G]:
        """Iterate over the value of the instance.

        Yields:
            The value of the instance.
        """
        yield self.value

    def iter(self) -> Iterator[G]:
        """Shortcut to transform to a list.

        Can be used as `list(x.iter())`. It will either contain a value or be
        an empty list.

        Yields:
            The value if the instance is a "Good" value.
        """
        if not self.is_bad:
            yield self.value

    def get_or_else(self, _default: LazyArg[G]) -> G:
        """Return the value.

        Args:
            _default: The default value in case the instance is "Bad".

        Returns:
            Either the value or the default specified by "default".
        """
        return self.value

    def map(self, fct: Callable[[G], K]) -> 'Good[B, K]':
        """Apply the function to its value if this is a `Good` instance.

        Args:
            fct: The function to apply to the "Good" value.

        Returns:
            Itself if its a `Bad` otherwise another instance of `Good`.
        """
        return Good(fct(self.value))

    def flat_map_bad(self, _fct: Callable[[B], OneOf[B, G]]) -> OneOf[B, G]:
        """Apply the input function if this is a `Bad` value.

        Args:
            _fct: The function to apply to the "Bad" value.

        Returns:
            Itself if its a `Good` otherwise another instance of `Bad`.
        """
        return self


OneOf = Bad[B, G] | Good[B, G]


@typing_extensions.deprecated(
    'The `is_bad` type guard is deprecated; use `isinstance(inst, Bad)` instead.',
)
def is_bad(inst: OneOf[B, G]) -> TypeGuard[Bad[B, G]]:
    """Assert that a OneOf instance is a `Bad`.

    Args:
        inst: The `OneOf` instance.

    Returns:
        True if the instance is a `Bad`.
    """
    # `m` does not reference this function anymore, excluding from coverage
    return isinstance(inst, Bad)  # pragma: no cover


@typing_extensions.deprecated(
    'The `is_good` type guard is deprecated; use `isinstance(inst, Good)` instead.',
)
def is_good(inst: OneOf[B, G]) -> TypeGuard[Good[B, G]]:
    """Assert that a OneOf instance is a `Good`.

    Args:
        inst: The `OneOf` instance.

    Returns:
        True if the instance is a `Good`.
    """
    # `m` does not reference this function anymore, excluding from coverage
    return isinstance(inst, Good)  # pragma: no cover



# noqa: WPS412
from .fp import Bad, Good, OneOf, is_bad
from .issue import Issue, IssueDict
from .one_of import hone, issue, non_issue, one_of
from .types import Res

__all__ = [  # noqa: WPS410
    'Issue',
    'IssueDict',
    'issue',
    'non_issue',
    'one_of',
    'Good',
    'OneOf',
    'Bad',
    'is_bad',
    'Res',
    'hone',
]



from m.cli import Arg, BaseModel, command, run_main, validate_payload
from m.core.io import env


class Arguments(BaseModel):
    r"""Create a pull request.

    https://docs.github.com/en/rest/reference/pulls#create-a-pull-request

    example::

        $ m github create_pr \
            --owner jmlopez-rod \
            --repo repo \
            --head feature_branch \
            --base master \
            --title 'PR Title' \
            @file_with_pr_body | m json
    """

    owner: str = Arg(
        default=env('GITHUB_REPOSITORY_OWNER'),
        help='repo owner',
    )
    repo: str = Arg(
        help='repo name',
        required=True,
    )
    head: str = Arg(
        help='name of the branch where the changes are implemented',
        required=True,
    )
    base: str = Arg(
        help='name of the branch you want the changes pulled into',
        required=True,
    )
    title: str = Arg(
        help='pull request title',
        required=True,
    )
    body: str = Arg(
        default='@-',
        help='data: @- (stdin), @filename (file), string',
        validator=validate_payload,
        positional=True,
    )


@command(
    help='create a pull request',
    model=Arguments,
)
def run(arg: Arguments, arg_ns) -> int:
    from m.github.api import GithubPullRequest, create_pr
    return run_main(lambda: create_pr(
        arg_ns.token,
        arg.owner,
        arg.repo,
        GithubPullRequest(
            title=arg.title,
            body=arg.body,
            head=arg.head,
            base=arg.base,
        ),
    ))



import json
import logging
import textwrap

from m.core import io as mio

from .ci_tools.ci_tools import get_ci_tool
from .ci_tools.types import Message


class JsonFormatter(logging.Formatter):
    """Format each record as JSON data in one single line."""

    def format(self, record: logging.LogRecord) -> str:
        """Format a record as json.

        Args:
            record: The logRecord to format.

        Returns:
            A formatted string.
        """
        record_dict = record.__dict__
        context = record_dict.get('context')
        ci_info = record_dict.get('ci_info', Message(msg=record.msg))
        replacements = {
            **record_dict,
            **ci_info.model_dump(),
            'asctime': self.formatTime(record, self.datefmt),
            'context': context,
        }
        replacements.pop('ci_info', '')
        sanitized = {
            key: non_empty
            for key, non_empty in replacements.items()
            if non_empty
        }
        return json.dumps(sanitized)


class CiFormatter(logging.Formatter):
    """A formatter that takes into account a CI environment.

    We can still use
    """

    def __init__(
        self,
        datefmt: str = '%I:%M:%S %p - %b %d, %Y',  # noqa: WPS323
    ):
        """Initialize by optionally providing a date formatter.

        See https://docs.python.org/3/library/time.html#time.strftime
        for more info on how to format the date.

        Args:
            datefmt: The date format to use.
        """
        super().__init__(datefmt=datefmt)
        self.opened_blocks: list[str] = []
        self.ci_tool = get_ci_tool()
        self.show_traceback = mio.is_traceback_enabled()
        self.debug_python = mio.is_python_info_enabled()

    def format(self, record: logging.LogRecord) -> str:
        """Format a record as based on the CI environment.

        Args:
            record: The logRecord to format.

        Returns:
            A formatted string.
        """
        msg = self.ci_tool.formatter(
            self,
            record,
            self.show_traceback,
            self.debug_python,
        )
        if self.ci_tool.ci:
            return msg

        record_dict = record.__dict__
        open_b = record_dict.get('open_block')
        close_b = record_dict.get('close_block')

        if open_b:
            name, _ = open_b
            self.opened_blocks.append(name)
        if close_b:
            while self.opened_blocks and self.opened_blocks[-1] != close_b:
                self.opened_blocks.pop()
            if self.opened_blocks:
                self.opened_blocks.pop()

        if not self.opened_blocks:
            return msg
        return textwrap.indent(msg, '  ' * len(self.opened_blocks))



from typing import cast

from pydantic.fields import FieldInfo

from ..misc import argument_description, argument_name
from ..types import MISSING, AnyMap, FuncArgs


def handle_field(name: str, field: FieldInfo) -> FuncArgs:
    """Treat the field as a boolean field.

    Args:
        name: The name of the argument.
        field: A dictionary representation of the field.

    Returns:
        Function arguments for the parser `add_argument` method.
    """
    default = field.default
    extras = cast(dict, field.json_schema_extra or {})
    aliases = cast(list[str], extras.get('aliases', None))
    args: AnyMap = {
        'help': argument_description(field.description or '', MISSING),
        'required': extras.get('required', False),
    }

    if default:
        args['action'] = 'store_false'
        args['dest'] = name
        names = [argument_name(f'no-{name}')]
        if aliases:
            names = [argument_name(f'no-{alias}') for alias in aliases]
    else:
        args['action'] = 'store_true'
        names = [argument_name(name)]
        if aliases:
            names = [argument_name(alias) for alias in aliases]

    return FuncArgs(args=names, kwargs=args)



# noqa: WPS412
from .conftest import run_action_step, run_action_test_case
from .testing import (
    ActionStepTestCase,
    block_m_side_effects,
    block_network_access,
    mock,
)

__all__ = [  # noqa: WPS410
    'ActionStepTestCase',
    'block_m_side_effects',
    'block_network_access',
    'mock',
    'run_action_step',
    'run_action_test_case',
]



# noqa: WPS412
from .actions import Action, KebabModel, RunStep, UsesStep
from .api import InArg, OutArg, run_action

__all__ = [  # noqa: WPS410
    'run_action',
    'KebabModel',
    'InArg',
    'OutArg',
    'Action',
    'RunStep',
    'UsesStep',
]



from typing import Callable, TypeVar

T = TypeVar('T')


def maybe(callback: Callable[[], T]) -> T | None:
    """Evaluate the callback to return a value.

    Unlike Typescript, Python does not have optional chaining::

        https://en.wikipedia.org/wiki/Safe_navigation_operator#Python

    To simulate this we can use this function as follows::

        ans = maybe(lambda: path.to.prop)  # type: ignore[union-attr]

    It is ok to disable the `union-attr` mypy check as long as mypy is checking
    for `no-any-return`.

    Args:
        callback: A function return a value.

    Returns:
        The value returned by the function or `None`.
    """
    try:
        return callback()
    except AttributeError:
        return None


def non_null(inst: T | None) -> T:
    """Assert that `inst` is not `None`.

    Implementation taken from::

        https://github.com/python/typing/issues/645#issuecomment-501057220

    Args:
        inst: A possibly null instance.

    Returns:
        The same argument with the `None` type removed.
    """
    # The assert statement can go away with the -O flag.
    assert inst is not None  # noqa: S101
    return inst



import os
import socket
from functools import partial
from typing import Any

from pydantic import BaseModel, Field


class BlockNetwork(socket.socket):
    """Implements a socket that blocks all network calls."""

    # pylint: disable-next=super-init-not-called
    def __init__(self, *args: Any, **kwargs: Any):
        """Raise an error if called.

        Args:
            args: ...
            kwargs: ...

        Raises:
            RuntimeError: Always.
        """
        raise RuntimeError('Network call blocked')


class ActionStepTestCase(BaseModel):
    """Defines a test case for an action.

    Useful in the parametrization of several use cases for an action.
    """

    name: str = Field(description="""
        Unique name for the test case. This is used as the identifier
        for a test case so that test failures may be easier to spot.
    """)

    py_file: str = Field(description='Path to the python file to run.')

    inputs: dict[str, str] = Field(description="""
        Inputs to the script. Should be of the form

        ```py
        {'INPUT_[SOME_NAME]': 'value'}
        ```
    """)

    exit_code: int = Field(
        default=0,
        description='The expected exit code (default: 0).',
    )

    expected_stdout: str = Field(
        default='',
        description='The expected stdout (default: empty).',
    )

    # Errors may be noisy, specify strings that are expected to be in stderr
    errors: list[str] = Field(
        default=[],
        description="""
            Errors may be noisy, specify strings that are expected to be in stderr
            in an array of strings.
        """,
    )

    # list of outputs: `output-name=output-value`
    outputs: list[str] = Field(
        default=[],
        description="""
            list of Github outputs. Each entry in the array should be of the form

            ```
            output-name=output-value
            ```
        """,
    )

    file_write_side_effect: Any | None = Field(
        default=None,
        description="""
            Defaults to [`Good(0)`][m.core.fp.Good]. This can be provided if we
            need to modify the behavior of [m.core.rw.write_file][]. This is
            useful if we want to test cases in which a file failed to write.
        """,
    )


def needs_mocking(func_name: str, *args: Any, **kwargs: Any):
    """Raise an exception asking developer to mock a function.

    Args:
        func_name: name of the function
        args: ...
        kwargs: ...

    Raises:
        RuntimeError: Always.
    """
    raise RuntimeError(f'DEV ERROR: Need to mock {func_name}({args},{kwargs})')


def mock(func_name: str) -> Any:
    """Create a function that raises an error if its not mocked.

    Args:
        func_name: full module path to the function to mock.

    Returns:
        A function that raises an error if its not mocked.
    """
    return partial(needs_mocking, func_name)


def block_network_access() -> None:
    """Blocks network access for all tests.

    This function overrides the definition of [`socket`][socket.socket] so that
    we do not accidentally try to run tests that make network calls. If our tests
    do not depend on other local services it is a good idea to call this before
    any of our tests runs.

    Otherwise we may want to modify this function to allow certain hosts to be
    called. (PRs welcomed).
    """
    # making sure that no calls to the internet are done
    socket.socket = BlockNetwork  # type: ignore


def block_m_side_effects() -> dict[str, Any]:
    """Blocks functions that have side effects.

    This function overrides the definition of `m` so that we do not accidentally
    try write a lot of files or create/move directories.

    The [pathlib.Path.mkdir][] function should only be blocked while developing
    tests. It is a reminder that we haven't mocked the function yet. If we want
    to get this reminder then add `touch m/.m/pytest-ran` after the tests run
    locally.

    Returns:
        A dictionary with references to the original functions that were
            overridden in case these are needed.
    """
    import shutil
    import subprocess  # noqa: S404 - importing to disable it during testing
    from pathlib import Path

    from m.core import rw as mio

    originals = {
        'write_file': mio.write_file,
        'Path_mkdir': Path.mkdir,
        'shutil_move': shutil.move,
    }

    mio.write_file = mock('m.core.rw.write_file')
    subprocess.check_output = mock('m.core.subprocess.eval_cmd')
    subprocess.call = mock('m.core.subprocess.exec_pnpm')
    shutil.move = mock('shutil.move')

    if not os.environ.get('CI'):
        # We want to make sure that we do not create directories during tests.
        # To do so we will mock the Path.mkdir function. There is a problem:
        # pytest needs this function to create directories for its own purposes.
        # For this reason we will only mock the function after we create the
        # m/.m/pytest-ran file.
        if Path('m/.m/pytest-ran').exists():
            Path.mkdir = mock('pathlib.Path.mkdir')  # type: ignore

    return originals



from textwrap import dedent
from typing import cast

from m.core import Bad, Good, Res, io, issue, one_of
from m.github.api import GithubPullRequest, create_pr
from m.github.ci import compare_sha_url
from m.github.cli import get_latest_release
from m.github.graphql.queries.branch_prs import PullRequest
from m.github.graphql.queries.branch_prs import fetch as fetch_branch_prs
from m.log import Logger

from m import git

from .config import Config, read_config
from .release_utils import YES_NO, assert_branch, is_yes

logger = Logger('m.ci.review_release')


def release_pr_body(config: Config, gh_ver: str) -> str:
    """Generate the pull request body.

    Args:
        config: The `m` configuration.
        gh_ver: The current version in Github.

    Returns:
        The text to add to the pull request.
    """
    link = compare_sha_url(config.owner, config.repo, gh_ver, 'HEAD')
    instructions = (
        '**DO NOT** use the merge button. Instead run `m end_release`.'
        if config.uses_git_flow()
        else 'either push the merge button or run `m end_release`'
    )
    return dedent(f"""\
        ## Reviewer directions

        Verify `CHANGELOG.md` contains a summary of the unreleased changes.

        {link}

        ## Author directions

        - Wait for reviewers to approve
        - When approved, {instructions}
    """)


def _git_flow_pr_body(config: Config, branch: str, gh_ver: str) -> str:
    link = compare_sha_url(config.owner, config.repo, gh_ver, 'HEAD')
    master = config.get_master_branch()
    develop = config.get_develop_branch()
    return dedent(f"""\
        ## Backport pull request

        This pull request attempts to merge commits created during the release
        process back to the `{develop}` branch. It should only be merged after
        the `{branch}` branch has been merged to `{master}`.

        ## Unreleased changes

        It may be easier to see the changelog in here. Please verify that
        the unreleased changes are properly documented in the `CHANGELOG`.

        {link}

        ## Merging

        This pull request should be merged via `m end_release` when this
        pull request and the "main" one are approved.
    """)


def acknowledge_git_status(status: str) -> Res[None]:
    """Display the current git status and ask developer to confirm.

    Args:
        status: The raw output of `git status`.

    Returns:
        An issue to stop the operation, otherwise None.
    """
    logger.info('The following changes will be committed', {'git': status})
    response = io.prompt_choices(
        'proceed creating the pull request(s)?',
        YES_NO,
        as_list=False,
    )
    if is_yes(response):
        return Good(None)
    return issue('operation cancelled by user')


def inspect_prs(all_prs: list[PullRequest]) -> Res[None]:
    """Inspect the release pull requests.

    There should not be any pull requests when calling `review_release`.
    This is a one time operation.

    Args:
        all_prs: The list of pull requests.

    Returns:
        An issue if prs already exist, None otherwise.
    """
    prs = [pr for pr in all_prs if pr.closed is False]
    if prs:
        return issue('release is already in review', context={
            'prs': {pr.number: pr.url for pr in prs},
        })
    return Good(None)


def create_prs(
    gh_token: str,
    config: Config,
    release_type: str,
    target_ver: str,
    gh_ver: str,
) -> Res[None]:
    """Create release pull request(s).

    Args:
        gh_token: The GITHUB_TOKEN to use to make api calls to Github.
        config: The m configuration.
        release_type: 'hotfix' or 'release'.
        target_ver: The version to release.
        gh_ver: The current version in Github.

    Returns:
        An issue if there is a problem while creating or None if successful.
    """
    all_prs: dict[str, str] = {}
    git_branch = f'{release_type}/{target_ver}'
    develop_branch = config.get_develop_branch()
    if config.uses_git_flow():
        title = f'({release_type} to {develop_branch}) {target_ver}'
        backport_pr = create_pr(
            gh_token,
            config.owner,
            config.repo,
            GithubPullRequest(
                title=title,
                body=_git_flow_pr_body(config, git_branch, gh_ver),
                head=git_branch,
                base=develop_branch,
            ),
        ).map(lambda res: cast(str, res.get('html_url', '')))
        if isinstance(backport_pr, Bad):
            logger.warning(
                'unable to create backport pull request',
                backport_pr.value,
            )
        else:
            all_prs[title] = backport_pr.value
    title = f'({release_type}) {target_ver}'
    release_pr = create_pr(
        gh_token,
        config.owner,
        config.repo,
        GithubPullRequest(
            title=title,
            body=release_pr_body(config, gh_ver),
            head=git_branch,
            base=config.get_master_branch(),
        ),
    ).map(lambda res: cast(str, res.get('html_url', '')))
    if isinstance(release_pr, Bad):
        logger.warning(
            'unable to create release pull request',
            release_pr.value,
        )
    else:
        all_prs[title] = release_pr.value
    if all_prs:
        logger.info('pull requests created', context=all_prs)
        return Good(None)
    return issue('no prs were created, inspect logs for hints')


def _commit_changes(commit_msg: str) -> Res[str]:
    commit_result = git.commit(commit_msg)
    if isinstance(commit_result, Bad):
        if 'working tree clean' in f'{commit_result.value}':
            return Good('')
    return commit_result


def review_release(token: str) -> Res[None]:
    """Create release prs.

    Args:
        token: The GITHUB_TOKEN to use to make api calls to Github.

    Returns:
        None if successful, otherwise an issue.
    """
    return one_of(lambda: [
        None
        for branch in git.get_branch()
        for release_type, target_ver in assert_branch(branch, 'review')
        for config in read_config('m')
        for prs in fetch_branch_prs(token, config.owner, config.repo, branch)
        for _ in inspect_prs(prs)
        for _ in git.stage_all()
        for git_status in git.raw_status()
        for _ in acknowledge_git_status(git_status)
        for _ in _commit_changes(f'({release_type}) {target_ver}')
        for _ in git.push_branch(branch)
        for gh_ver in get_latest_release(token, config.owner, config.repo)
        for _ in create_prs(token, config, release_type, target_ver, gh_ver)
    ])
